import logging

from ios_device.util.lockdown import LockdownClient


class SBServiceClient(object):
    SERVICE_NAME = "com.apple.springboardservices"

    def __init__(self, lockdown=None, udid=None, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.lockdown = lockdown or LockdownClient(udid=udid)
        self.service = self.lockdown.start_service(self.SERVICE_NAME)

    def get_icon_state(self, format_version="2"):
        cmd = {"command": "getIconState"}
        if format_version:
            cmd["formatVersion"] = format_version

        self.service.send_plist(cmd)
        res = self.service.recv_plist()
        return res

    def set_icon_state(self, newstate={}):
        cmd = {"command": "setIconState",
               "iconState": newstate}

        self.service.send_plist(cmd)

    def get_icon_pngdata(self, bid):
        cmd = {"command": "getIconPNGData",
               "bundleId": bid}

        self.service.send_plist(cmd)
        res = self.service.recv_plist()
        pngdata = res.get("pngData")
        if res:
            return pngdata
        return None

    def get_interface_orientation(self):
        cmd = {"command": "getInterfaceOrientation"}
        self.service.send_plist(cmd)
        res = self.service.recv_plist()
        return res.get('interfaceOrientation')

    def get_wallpaper_pngdata(self):
        cmd = {"command": "getHomeScreenWallpaperPNGData"}
        self.service.send_plist(cmd)
        res = self.service.recv_plist()
        if res:
            return res.get("pngData")
        return None



import time
import uuid

from ios_device.cli.base import InstrumentDeviceInfo
from ios_device.servers.Instrument import InstrumentServer
from ios_device.util.lifecycle import AppLifeCycle
from ios_device.util.dtx_msg import RawObj
from ios_device.util.exceptions import InstrumentRPCParseError
from ios_device.util.kperf_data import KperfData, KdBufParser


def app_launch_lifecyle(rpc, bundleid):
    traceCodesFile = InstrumentDeviceInfo(rpc).traceCodesFile()
    Kperf = KperfData(traceCodesFile)
    machTimeInfo = rpc.call("com.apple.instruments.server.services.deviceinfo", "machTimeInfo").selector
    usecs_since_epoch = rpc.lockdown.get_value(key='TimeIntervalSince1970') * 1000000
    LifeCycle = AppLifeCycle(machTimeInfo, usecs_since_epoch)

    def demo(data):
        for event in Kperf.to_dict(data):
            if isinstance(event, KdBufParser):
                _, process_name = Kperf.threads_pids.get(event.tid, (None, None))
                if event.class_code in (0x1f, 0x2b, 0x31) and process_name not in ('SpringBoard',):
                    LifeCycle.decode_app_lifecycle(event, process_name or event.tid)
                if event.debug_id == 835321862:
                    LifeCycle.format_str()

    def on_graphics_message(res):
        if type(res.selector) is InstrumentRPCParseError:
            demo(res.selector.data)

    rpc.register_channel_callback("com.apple.instruments.server.services.coreprofilesessiontap", on_graphics_message)
    rpc.call("com.apple.instruments.server.services.coreprofilesessiontap", "setConfig:",
             {'rp': 100,
              'bm': 1,
              'tc': [{'kdf2': {735576064, 835321856, 735838208, 730267648,
                               520552448},
                      'csd': 128,
                      'tk': 3,
                      'ta': [[3], [0], [2], [1, 1, 0]],
                      'uuid': str(uuid.uuid4()).upper()}],
              })
    rpc.call("com.apple.instruments.server.services.coreprofilesessiontap", "start")
    channel = "com.apple.instruments.server.services.processcontrol"
    pid = rpc.call(channel, 'launchSuspendedProcessWithDevicePath:bundleIdentifier:environment:arguments:options:',
                    '',
                    bundleid,
                    {'OS_ACTIVITY_DT_MODE': '1', 'HIPreventRefEncoding': '1', 'DYLD_PRINT_TO_STDERR': '1'}, [],
                    {'StartSuspendedKey': 0}).selector
    print(f'start {bundleid} pid:{pid}')

    time.sleep(1000)
    rpc.call("com.apple.instruments.server.services.coreprofilesessiontap", "stop")
    rpc.stop()


if __name__ == '__main__':
    rpc = InstrumentServer().init()
    app_launch_lifecyle(rpc, 'cn.rongcloud.im')
    rpc.stop()



#!/usr/bin/env python3
import logging

from ios_device.util.lockdown import LockdownClient


class AmfiService:
    SERVICE_NAME = 'com.apple.amfi.lockdown'

    def __init__(self, lockdown=None, udid=None, network=None, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.lockdown = lockdown or LockdownClient(udid=udid, network=network)

    def enable_developer_mode(self):
        """ The device will restart and enable Developer mode
        :return:
        """
        service = self.lockdown.start_service(self.SERVICE_NAME)
        resp = service.plist_request({'action': 1})
        if not resp.get('success'):
            raise Exception('enable developer mode error')
        self.logger.info('device will restart ...')

    def enable_developer_mode_turn_on(self):
        """ answer the prompt that appears after the restart with "turn on" """
        service = self.lockdown.start_service(self.SERVICE_NAME)
        resp = service.plist_request({'action': 2})
        if not resp.get('success'):
            raise Exception(f'enable_developer_mode_turn_on {resp}')


if __name__ == '__main__':
    AmfiService().enable_developer_mode()
    # AmfiService().enable_developer_mode_turn_on()



"""
@Date    : 2021-01-28
@Author  : liyachao
"""
import logging
import os
import struct
import threading
from distutils.version import LooseVersion

from ios_device.util.bpylist2 import NSUUID, XCTestConfiguration, NSURL

from ios_device.servers.Installation import InstallationProxyService
from ios_device.servers.Instrument import InstrumentServer
from ios_device.servers.dvt import DTXEnum
from ios_device.servers.house_arrest import HouseArrestService
from ios_device.servers.testmanagerd import TestManagerdLockdown
from ios_device.util.bpylist2 import archive
from ios_device.util.dtx_msg import RawObj
from ios_device.util.forward import ForwardPorts
from ios_device.util.lockdown import LockdownClient


def channel_validate(channel: InstrumentServer):
    if not channel:
        return False, "The rpc channel is not enabled. Please initialize it first"
    return True, ""


def caller(res, func):
    """
    异步回调中转器
    :param res:
    :param func:
    :return:
    """
    if not isinstance(res, bytes):
        res_dic = dict()
        header = res.selector
        if header:
            res_dic["header"] = header
        body = res.auxiliaries
        if body:
            res_dic["body"] = body
        func(res_dic)
    else:
        func(res)


def network_caller(res, func):
    """
    网络异步回调包解析
    :param res:
    :param func:
    :return:
    """
    from construct import Struct, Int32ul, Bytes, Adapter, Switch, this, Int16ub, Int8ul
    import ipaddress

    class IP(Adapter):
        def _decode(self, obj, context, path):
            return ipaddress.ip_address(obj)

    headers = {
        0: ['InterfaceIndex', "Name"],
        1: ['LocalAddress', 'RemoteAddress', 'InterfaceIndex', 'Pid', 'RecvBufferSize', 'RecvBufferUsed',
            'SerialNumber', 'Kind'],
        2: ['RxPackets', 'RxBytes', 'TxPackets', 'TxBytes', 'RxDups', 'RxOOO', 'TxRetx', 'MinRTT', 'AvgRTT',
            'ConnectionSerial']
    }
    msg_type = {
        0: "interface-detection",
        1: "connection-detected",
        2: "connection-update",
    }

    SockAddr = Struct(
        'len' / Int8ul,
        Int8ul,
        'port' / Int16ub,
        'network' / Switch(this.len, {
            0x1c: Struct(
                Int32ul,
                'address' / IP(Bytes(16)),
                Int32ul, ),
            0x10: Struct(
                'address' / IP(Bytes(4)),
                Bytes(8))})
    )
    data = res.selector
    if data[0] == 1:
        addr = SockAddr.parse(data[1][0])
        data[1][0] = f"{addr.network.address}:{addr.port}"
        addr = SockAddr.parse(data[1][1])
        data[1][1] = f"{addr.network.address}:{addr.port}"

    func({str(msg_type[data[0]]): dict(zip(headers[data[0]], data[1]))})


def system_caller(res, func):
    if isinstance(res.selector, list):
        func(res.selector)


class PyIOSDeviceException(Exception):

    def __init__(self, *args):
        self.args = args


class RunXCUITest(threading.Thread):
    def __init__(self, bundle_id, callback, device_id=None, app_env: dict = None,
                 pair_ports=None):
        super().__init__()
        if pair_ports is None:
            pair_ports = ["8100:8100"]
        self.device_id = device_id
        self.bundle_id = bundle_id
        self.quit_event = threading.Event()
        self.callback = callback
        self.app_env = app_env
        self.lock_down = None
        self.forward = False
        self.pair_ports = pair_ports
        self.forward_thread = None

    def stop(self):
        if self.forward and self.forward_thread:
            self.forward_thread.stop()
        self.quit_event.set()

    def run(self) -> None:
        def _callback(res):
            self.callback(res.auxiliaries)

        lock_down = LockdownClient(udid=self.device_id)
        installation = InstallationProxyService(lockdown=lock_down)

        if self.pair_ports:
            self.forward = True
            self.forward_thread = ForwardPorts(pair_ports=self.pair_ports, device_id=self.device_id)
            self.forward_thread.start()

        app_info = installation.find_bundle_id(self.bundle_id)
        if not app_info:
            raise Exception("No app matches", self.bundle_id)
        logging.debug("BundleID: %s", self.bundle_id)
        logging.debug("DeviceIdentifier: %s", lock_down.device_info.get('UniqueDeviceID'))
        sign_identity = app_info.get("SignerIdentity", "")
        logging.debug("SignIdentity: %r", sign_identity)
        xcode_version = 29
        session_identifier = NSUUID(bytes=os.urandom(16), version=4)
        manager_lock_down_1 = TestManagerdLockdown(lock_down).init()

        manager_lock_down_1.make_channel("dtxproxy:XCTestManager_IDEInterface:XCTestManager_DaemonConnectionInterface")
        if lock_down.ios_version > LooseVersion('11.0'):
            result = manager_lock_down_1.call(
                "dtxproxy:XCTestManager_IDEInterface:XCTestManager_DaemonConnectionInterface",
                "_IDE_initiateControlSessionWithProtocolVersion:", RawObj(xcode_version)).selector
            logging.debug("result: %s", result)
        manager_lock_down_1.register_selector_callback(DTXEnum.FINISHED, lambda _: self.quit_event.set())
        manager_lock_down_1.register_undefined_callback(_callback)

        manager_lock_down_2 = TestManagerdLockdown(lock_down).init()
        manager_lock_down_2.make_channel("dtxproxy:XCTestManager_IDEInterface:XCTestManager_DaemonConnectionInterface")
        manager_lock_down_2.register_selector_callback(DTXEnum.FINISHED, lambda _: self.quit_event.set())
        manager_lock_down_2.register_undefined_callback(_callback)

        _start_flag = threading.Event()

        def _start_executing(res=None):
            if _start_flag.is_set():
                return
            _start_flag.set()

            logging.debug(" _start_executing Start execute test plan with IDE version: %d",
                          xcode_version)
            manager_lock_down_2._call(False, 0xFFFFFFFF, '_IDE_startExecutingTestPlanWithProtocolVersion:',
                                      RawObj(xcode_version))

        def _show_log_message(res):
            # logging.info(f"{res.selector} : {get_auxiliary_text(res.raw)}")
            if 'Received test runner ready reply with error: (null' in ''.join(
                    res.auxiliaries):
                logging.debug("_start_executing Test runner ready detected")
                _start_executing()

        manager_lock_down_2.register_selector_callback('_XCT_testBundleReadyWithProtocolVersion:minimumVersion:',
                                                       _start_executing)
        manager_lock_down_2.register_selector_callback('_XCT_logDebugMessage:', _show_log_message)
        manager_lock_down_2.register_selector_callback('_XCT_didFinishExecutingTestPlan',
                                                       lambda _: self.quit_event.set())

        result = manager_lock_down_2.call('dtxproxy:XCTestManager_IDEInterface:XCTestManager_DaemonConnectionInterface',
                                          '_IDE_initiateSessionWithIdentifier:forClient:atPath:protocolVersion:',
                                          RawObj(
                                              session_identifier,
                                              str(session_identifier) + '-6722-000247F15966B083',
                                              '/Applications/Xcode.app/Contents/Developer/usr/bin/xcodebuild',
                                              xcode_version
                                          )).selector
        logging.debug("_start_executing result: %s", result)
        # launch_wda
        xctest_path = "/tmp/WebDriverAgentRunner-" + str(session_identifier).upper() + ".xctestconfiguration"
        xctest_content = archive(XCTestConfiguration({
            "testBundleURL": NSURL(None, "file://" + app_info['Path'] + "/PlugIns/WebDriverAgentRunner.xctest"),
            "sessionIdentifier": session_identifier,
        }))

        fsync = HouseArrestService(lock_down)
        fsync.send_command(self.bundle_id)
        for fname in fsync.read_directory("/tmp"):
            if fname.endswith(".xctestconfiguration"):
                logging.debug("remove /tmp/%s", fname)
                fsync.file_remove("/tmp/" + fname)
        fsync.set_file_contents(xctest_path, xctest_content)

        conn = InstrumentServer(lock_down).init()
        conn.call('com.apple.instruments.server.services.processcontrol', 'processIdentifierForBundleIdentifier:',
                  self.bundle_id)

        conn.register_undefined_callback(_callback)
        app_path = app_info['Path']
        app_container = app_info['Container']

        xctestconfiguration_path = app_container + xctest_path
        logging.debug("AppPath: %s", app_path)
        logging.debug("AppContainer: %s", app_container)

        app_env = {
            'CA_ASSERT_MAIN_THREAD_TRANSACTIONS': '0',
            'CA_DEBUG_TRANSACTIONS': '0',
            'DYLD_FRAMEWORK_PATH': app_path + '/Frameworks:',
            'DYLD_LIBRARY_PATH': app_path + '/Frameworks',
            'NSUnbufferedIO': 'YES',
            'SQLITE_ENABLE_THREAD_ASSERTIONS': '1',
            'WDA_PRODUCT_BUNDLE_IDENTIFIER': '',
            'XCTestConfigurationFilePath': xctestconfiguration_path,
            'XCODE_DBG_XPC_EXCLUSIONS': 'com.apple.dt.xctestSymbolicator',
            'MJPEG_SERVER_PORT': '',
            'USE_PORT': '',
        }
        if self.app_env:
            app_env.update(self.app_env)
        if lock_down.ios_version > LooseVersion('11.0'):
            app_env['DYLD_INSERT_LIBRARIES'] = '/Developer/usr/lib/libMainThreadChecker.dylib'
            app_env['OS_ACTIVITY_DT_MODE'] = 'YES'
        app_options = {'StartSuspendedKey': False}
        if lock_down.ios_version > LooseVersion('12.0'):
            app_options['ActivateSuspended'] = True

        app_args = [
            '-NSTreatUnknownArgumentsAsOpen', 'NO',
            '-ApplePersistenceIgnoreState', 'YES'
        ]

        identifier = "launchSuspendedProcessWithDevicePath:bundleIdentifier:environment:arguments:options:"

        pid = conn.call('com.apple.instruments.server.services.processcontrol', identifier,
                        app_path, self.bundle_id, app_env, app_args, app_options).selector
        if not isinstance(pid, int):
            logging.error(f"Launch failed: {pid}")
            raise Exception("Launch failed")

        logging.debug(f" Launch {self.bundle_id} pid: {pid}")

        conn.call('com.apple.instruments.server.services.processcontrol', "startObservingPid:", RawObj(pid))

        if self.quit_event:
            conn.register_selector_callback(DTXEnum.FINISHED, lambda _: self.quit_event.set())

        if lock_down.ios_version > LooseVersion('12.0'):
            identifier = '_IDE_authorizeTestSessionWithProcessID:'
            result = manager_lock_down_1.call(
                'dtxproxy:XCTestManager_IDEInterface:XCTestManager_DaemonConnectionInterface',
                identifier,
                RawObj(pid)).selector
            logging.debug("_IDE_authorizeTestSessionWithProcessID: %s", result)
        else:
            identifier = '_IDE_initiateControlSessionForTestProcessID:protocolVersion:'
            result = manager_lock_down_1.call(
                'dtxproxy:XCTestManager_IDEInterface:XCTestManager_DaemonConnectionInterface',
                identifier,
                RawObj(pid, xcode_version)).selector
            logging.debug("_IDE_authorizeTestSessionWithProcessID: %s", result)

        while not self.quit_event.wait(.1):
            pass
        logging.warning("xctrunner quited")
        conn.stop()
        manager_lock_down_2.stop()
        manager_lock_down_1.stop()



"""
获取系统相关信息，类似 Android TOP，包含进程信息，需要 iOS > 11
"""
import os
import sys
sys.path.append(os.getcwd())

from ios_device.servers.Instrument import InstrumentServer

import json
import time
from ios_device.util import logging

log = logging.getLogger(__name__)


def sysmontap(rpc):

    def dropped_message(res):
        print("[DROP]", res.selector, res.raw.channel_code)

    def on_sysmontap_message(res):
        if isinstance(res.selector, list):
            print(json.dumps(res.selector, indent=4))

    rpc.register_undefined_callback(dropped_message)
    rpc.call("com.apple.instruments.server.services.sysmontap", "setConfig:", {
        'ur': 1000,  # 输出频率 ms
        'bm': 0,
        'procAttrs': ['memVirtualSize', 'cpuUsage', 'procStatus', 'appSleep', 'uid', 'vmPageIns', 'memRShrd',
                      'ctxSwitch', 'memCompressed', 'intWakeups', 'cpuTotalSystem', 'responsiblePID', 'physFootprint',
                      'cpuTotalUser', 'sysCallsUnix', 'memResidentSize', 'sysCallsMach', 'memPurgeable',
                      'diskBytesRead', 'machPortCount', '__suddenTerm', '__arch', 'memRPrvt', 'msgSent', 'ppid',
                      'threadCount', 'memAnon', 'diskBytesWritten', 'pgid', 'faults', 'msgRecv', '__restricted', 'pid',
                      '__sandbox'],  # 输出所有进程信息字段，字段顺序与自定义相同（全量自字段，按需使用）
        'sysAttrs': ['diskWriteOps', 'diskBytesRead', 'diskBytesWritten', 'threadCount', 'vmCompressorPageCount',
                     'vmExtPageCount', 'vmFreeCount', 'vmIntPageCount', 'vmPurgeableCount', 'netPacketsIn',
                     'vmWireCount', 'netBytesIn', 'netPacketsOut', 'diskReadOps', 'vmUsedCount', '__vmSwapUsage',
                     'netBytesOut'],  # 系统信息字段
        'cpuUsage': True,
        'sampleInterval': 1000000000})
    rpc.register_channel_callback("com.apple.instruments.server.services.sysmontap", on_sysmontap_message)
    var = rpc.call("com.apple.instruments.server.services.sysmontap", "start").selector
    print(f"start {var}")
    time.sleep(1000)
    var = rpc.call("com.apple.instruments.server.services.sysmontap", "stop").selector
    print(f"stop {var}")
    rpc.stop()


if __name__ == '__main__':
    # rpc = InstrumentServer()
    # addresses, port, psk = rpc.start_wireless()
    # print('start wireless', addresses, port, psk)
    # rpc = rpc.init_wireless(addresses, port, psk)
    rpc = InstrumentServer().init()
    sysmontap(rpc)
    rpc.stop()



import math
from collections import namedtuple, defaultdict

from ios_device.util.kperf_data import KdBufParser

AppLifeEvent = namedtuple('AppLifeEvent', ['period', 'sub_state', 'time', 'thread', 'kind', 'code'])


def convertTime(data):
    lst = ['ns', 'us', 'ms']
    i = 0
    if data:
        i = int(math.floor(
            math.log(data, 1000)
        ))
    if i >= len(lst):
        i = len(lst) - 1
    return ('%.2f' + " " + lst[i]) % (data / math.pow(1000, i))


class AppLifeCycle:

    def __init__(self, machTimeInfo, usecs_since_epoch):
        self.mach_absolute_time = machTimeInfo[0]
        self.numer = machTimeInfo[1]
        self.denom = machTimeInfo[2]
        self.usecs_since_epoch = usecs_since_epoch
        self.events = defaultdict(list)
        self.process = None
        self.main_ui_thread = None

    def app_launching(self, process):
        self.process = process

    def update_app_period(self, data: AppLifeEvent):
        if self.events[data.thread]:
            self.events[data.thread].append(data)

    def update_start_period(self, data):
        self.events[data.thread].append(data)

    def format_timestamp(self, timestamp):
        return ((timestamp - self.mach_absolute_time) * self.numer) / self.denom

    def format_str(self):
        total_time = 0
        _tmp_sub_state = {}
        for key, events in self.events.items():
            if events and events[-1].sub_state == 'Initial Frame Rendering':
                for i in events:
                    if i.kind == 'BEGIN':
                        _tmp_sub_state[i.sub_state] = self.format_timestamp(i.time)
                    elif i.kind == 'END':
                        if i.sub_state in _tmp_sub_state:
                            end_time = self.format_timestamp(i.time)
                            _tmp_time = end_time - _tmp_sub_state[i.sub_state]
                            total_time += _tmp_time
                            print(f'{convertTime(_tmp_time):>10}   {i.period}-{i.sub_state}')
                            del _tmp_sub_state[i.sub_state]
                print(f'App Thread Process ID:{key}, Total Time:{convertTime(total_time)}')

                self.events[key].clear()

    def decode_app_lifecycle(self, event: KdBufParser, thread):
        if event.class_code == 0x1f:  # dyld-init
            if event.subclass_code == 0x7 and event.final_code == 13:
                self.update_start_period(AppLifeEvent('Initializing',
                                                      'System Interface Initialization (Dyld init)', event.timestamp,
                                                      thread, 'BEGIN',
                                                      event.debug_id))
            elif event.subclass_code == 0x7 and event.final_code == 1 and event.func_code == 2:
                self.update_app_period(AppLifeEvent('Initializing',
                                                    'Static Runtime Initialization', event.timestamp, thread, 'END',
                                                    event.debug_id))

        elif event.class_code == 0x31 and event.subclass_code == 0xca and event.final_code == 1 and event.func_code == 2:  # AppKit/UIKit common application launch phases
            self.update_app_period(AppLifeEvent('Launching',
                                                'Initial Frame Rendering', event.timestamp, thread, 'END',
                                                event.debug_id))

        elif event.class_code == 0x2b:
            if event.subclass_code == 0xd8:  # appkit-init
                if event.final_code == 1 and event.func_code == 0:
                    if not self.main_ui_thread:
                        self.update_app_period(AppLifeEvent('Launching',
                                                            'AppKit Initialization', event.timestamp, thread,
                                                            'BEGIN', event.debug_id))
                        self.main_ui_thread = 'UIKIT'
                    elif self.main_ui_thread == 'UIKIT':
                        self.update_app_period(AppLifeEvent('Launching',
                                                            'UIKit Initialization', event.timestamp, thread, 'END',
                                                            event.debug_id))
                        self.update_app_period(AppLifeEvent('Launching',
                                                            'AppKit Initialization', event.timestamp, thread, 'BEGIN',
                                                            event.debug_id))
                        self.main_ui_thread = 'MARZIPAN'

                elif event.final_code == 12 and event.func_code == 0:
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'AppKit Initialization', event.timestamp, thread, 'END',
                                                        event.debug_id))
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'AppKit Scene Creation', event.timestamp, thread, 'BEGIN',
                                                        event.debug_id))
                elif event.final_code == 12 and event.func_code == 1:
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'AppKit Scene Creation', event.timestamp, thread, 'END',
                                                        event.debug_id))
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'applicationWillFinishLaunching()', event.timestamp, thread,
                                                        'BEGIN', event.debug_id))
                elif event.final_code == 12 and event.func_code == 2:
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'applicationWillFinishLaunching()', event.timestamp, thread,
                                                        'END', event.debug_id))
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'AppKit Scene Creation', event.timestamp, thread, 'BEGIN',
                                                        event.debug_id))
                elif event.final_code == 11 and event.func_code == 1:
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'AppKit Scene Creation', event.timestamp, thread, 'END',
                                                        event.debug_id))
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'applicationDidFinishLaunching()', event.timestamp, thread,
                                                        'BEGIN', event.debug_id))

                elif event.final_code == 11 and event.func_code == 2:
                    if self.main_ui_thread == 'APPKIT':
                        self.update_app_period(AppLifeEvent('Launching',
                                                            'applicationDidFinishLaunching()', event.timestamp, thread,
                                                            'END', event.debug_id))
                        self.update_app_period(AppLifeEvent('Launching',
                                                            'Initial Frame Rendering', event.timestamp, thread, 'BEGIN',
                                                            event.debug_id))
                    elif self.main_ui_thread == 'MARZIPAN':
                        self.update_app_period(AppLifeEvent('Launching',
                                                            'applicationDidFinishLaunching()', event.timestamp, thread,
                                                            'END', event.debug_id))
                        self.update_app_period(AppLifeEvent('Launching',
                                                            'AppKit Scene Creation', event.timestamp, thread, 'BEGIN',
                                                            event.debug_id))

            elif event.subclass_code == 0x87:  # UIKit application launch phases
                if event.final_code == 90 and event.args[0] == 0x32:
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'UIKit Initialization', event.timestamp, thread, 'BEGIN',
                                                        event.debug_id))
                    self.main_ui_thread = 'UIKIT'

                elif event.final_code == 21:
                    if self.main_ui_thread == 'UIKIT':
                        self.update_app_period(AppLifeEvent('Launching',
                                                            'UIKit Initialization', event.timestamp, thread, 'END',
                                                            event.debug_id))
                        self.update_app_period(AppLifeEvent('Launching',
                                                            'UIKit Scene Creation', event.timestamp, thread, 'BEGIN',
                                                            event.debug_id))
                    elif self.main_ui_thread == 'MARZIPAN':

                        self.update_app_period(AppLifeEvent('Initializing',
                                                            'AppKit Scene Creation', event.timestamp, thread, 'END',
                                                            event.debug_id))
                        self.update_app_period(AppLifeEvent('Initializing',
                                                            'UIKit Scene Creation', event.timestamp, thread, 'BEGIN',
                                                            event.debug_id))

                elif event.final_code == 23:
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'UIKit Scene Creation', event.timestamp, thread, 'END',
                                                        event.debug_id))
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'willFinishLaunchingWithOptions()', event.timestamp, thread,
                                                        'BEGIN', event.debug_id))
                elif event.final_code == 24:
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'willFinishLaunchingWithOptions()', event.timestamp, thread,
                                                        'END', event.debug_id))
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'UIKit Scene Creation', event.timestamp, thread, 'BEGIN',
                                                        event.debug_id))

                elif event.final_code == 25:
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'UIKit Scene Creation', event.timestamp, thread, 'END',
                                                        event.debug_id))
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'didFinishLaunchingWithOptions()', event.timestamp, thread,
                                                        'BEGIN', event.debug_id))

                elif event.final_code == 26:
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'didFinishLaunchingWithOptions()', event.timestamp, thread,
                                                        'END', event.debug_id))
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'UIKit Scene Creation', event.timestamp, thread, 'BEGIN',
                                                        event.debug_id))

                elif event.final_code == 300:
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'UIKit Scene Creation', event.timestamp, thread, 'END',
                                                        event.debug_id))
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'sceneWillConnectTo()', event.timestamp, thread, 'BEGIN',
                                                        event.debug_id))
                elif event.final_code == 301:
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'sceneWillConnectTo()', event.timestamp, thread, 'END',
                                                        event.debug_id))
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'UIKit Scene Creation', event.timestamp, thread, 'BEGIN',
                                                        event.debug_id))
                elif event.final_code == 312:
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'UIKit Scene Creation', event.timestamp, thread, 'END',
                                                        event.debug_id))
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'sceneWillEnterForeground()', event.timestamp, thread, 'BEGIN',
                                                        event.debug_id))

                elif event.final_code == 313:
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'sceneWillEnterForeground()', event.timestamp, thread, 'END',
                                                        event.debug_id))
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'UIKit Scene Creation', event.timestamp, thread, 'BEGIN',
                                                        event.debug_id))

                elif event.final_code == 12:
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'UIKit Scene Creation', event.timestamp, thread, 'END',
                                                        event.debug_id))
                    self.update_app_period(AppLifeEvent('Launching',
                                                        'Initial Frame Rendering', event.timestamp, thread, 'BEGIN',
                                                        event.debug_id))
            elif event.subclass_code == 0xdc:  # appkit-init
                if event.final_code == 4 and event.func_code == 0 and event.args[0] == 10:
                    self.update_app_period(AppLifeEvent('Initializing',
                                                        'System Interface Initialization (Dyld init)', event.timestamp,
                                                        thread,
                                                        'END', event.debug_id))
                    self.update_app_period(AppLifeEvent('Initializing',
                                                        'Static Runtime Initialization', event.timestamp, thread,
                                                        'BEGIN', event.debug_id))



"""
graphics 计算 fps
"""

import time
from ios_device.servers.Instrument import InstrumentServer


def cmd_graphics(rpc):
    last_timestamp = 0

    def dropped_message(res):
        print("[DROP]", res.selector, res.raw.channel_code)

    def on_graphics_message(res):
        data = res.selector
        nonlocal last_timestamp
        cost_time = data['XRVideoCardRunTimeStamp'] - last_timestamp
        last_timestamp = data['XRVideoCardRunTimeStamp']
        print(cost_time,'fps:', data['CoreAnimationFramesPerSecond'])

    rpc.register_undefined_callback(dropped_message)
    rpc.register_channel_callback("com.apple.instruments.server.services.graphics.opengl", on_graphics_message)
    print("start",
          rpc.call("com.apple.instruments.server.services.graphics.opengl", "startSamplingAtTimeInterval:", 0.0).selector)
    time.sleep(100)
    print("stop", rpc.call("com.apple.instruments.server.services.graphics.opengl", "stopSampling").selector)
    rpc.stop()


if __name__ == '__main__':
    rpc = InstrumentServer().init()
    cmd_graphics(rpc)
    rpc.stop()



import contextlib
import os
import plistlib
import shutil
import sys
import tempfile
import time
import uuid
import platform

import zipfile

from distutils.version import LooseVersion
from pathlib import Path
from typing import Optional, Dict, Any, Mapping

import requests

from .exceptions import PairingError, NotTrustedError, FatalPairingError, NotPairedError, CannotStopSessionError
from .exceptions import StartServiceError, InitializationError
from .plist_service import PlistService
from .ca import make_certs_and_key
from .usbmux import MuxDevice, UsbmuxdClient
from .utils import DictAttrProperty, cached_property
from .variables import LOG
from ..servers.image_mounter import MobileImageMounter
from ..util import Log, PROGRAM_NAME

__all__ = ['LockdownClient']
log = Log.getLogger(LOG.LockDown.value)
DEVICE_SUPPORT_SIZE = 5 * 1024
DOWN_SUPPORT_TIMEOUT = 15


def get_app_dir(*paths) -> str:
    home = os.path.expanduser("~")
    appdir = os.path.join(home, "." + PROGRAM_NAME)
    if paths:
        appdir = os.path.join(appdir, *paths)
    os.makedirs(appdir, exist_ok=True)
    return appdir


class LockdownClient:
    label = 'pyMobileDevice'
    udid = DictAttrProperty('device_info', 'UniqueDeviceID')
    unique_chip_id = DictAttrProperty('device_info', 'UniqueChipID')
    device_id = DictAttrProperty('device_info', 'DeviceID')
    ios_version = DictAttrProperty('device_info', 'ProductVersion', LooseVersion)

    def __init__(
            self,
            udid: Optional[str] = None,
            device: Optional[MuxDevice] = None,
            cache_dir: str = '.cache/pymobiledevice',
            network=None
    ):
        self.network = network
        self.cache_dir = cache_dir
        self.record = None  # type: Optional[Dict[str, Any]]
        self.sslfile = None
        self.session_id = None
        self.host_id = str(uuid.uuid3(uuid.NAMESPACE_DNS, platform.node())).upper()
        self.svc = PlistService(62078, udid, device, network=network)
        self.udid = self.svc.device.serial
        self.device = device
        self.network = network
        self._verify_query_type()
        self.device_info = self.get_value()
        self.device_info['UniqueDeviceID'] = self.svc.device.serial
        self.device_info['DeviceID'] = self.svc.device.device_id
        self.paired = None
        # log.info(f"Connecting Device {self.svc.device.serial} ")

    def _verify_query_type(self):
        query_type = self.svc.plist_request({'Request': 'QueryType'}).get('Type')
        if query_type != 'com.apple.mobile.lockdown':
            raise InitializationError(f'Unexpected {query_type}')

    @cached_property
    def identifier(self):
        if self.udid:
            return self.udid
        elif self.unique_chip_id:
            return f'{self.unique_chip_id:x}'
        raise InitializationError('Unable to determine UDID or ECID - failing')

    def _pair(self):
        if self._validate_pairing():
            return True
        self.pair_full()
        self.svc.close()
        self.svc = PlistService(62078, self.udid, self.svc.device, network=self.network)
        if self._validate_pairing():
            return True
        raise FatalPairingError

    def _get_pair_record(self) -> Optional[Dict[str, Any]]:
        lockdown_path = _get_lockdown_dir()
        itunes_lockdown_path = lockdown_path.joinpath(f'{self.identifier}.plist')
        try:  # 如果没有 lockdown 权限，则使用自有缓存证书，建议开启 lockdown 权限，避免重复认证
            if itunes_lockdown_path.exists():
                log.debug(f'Using iTunes pair record: {itunes_lockdown_path}')
                with itunes_lockdown_path.open('rb') as f:
                    return plistlib.load(f)
        except Exception as E:
            log.debug(f'{E}')
            log.debug(f'No iTunes pairing record found for device {self.identifier}')
            log.debug('Getting pair record from usbmuxd')
        with UsbmuxdClient() as usb:
            return usb.get_pair_record(self.udid)


    def _validate_pairing(self):
        pair_record = self._get_pair_record() or {}
        self.record = pair_record
        if self.ios_version < LooseVersion('11.0'):  # 11 以下需要双向认证
            resp = self._plist_request('ValidatePair', PairRecord=pair_record)
            if not resp or 'Error' in resp:
                log.error(f'Failed to ValidatePair: {resp}')
                return False

        self.host_id = pair_record.get('HostID', self.host_id)
        system_buid = pair_record.get('SystemBUID') or str(uuid.uuid3(uuid.NAMESPACE_DNS, platform.node())).upper()
        resp = self._plist_request('StartSession', HostID=self.host_id, SystemBUID=system_buid)
        self.session_id = resp.get('SessionID')

        if 'Error' in resp:
            if resp['Error'] == 'InvalidHostID':
                with UsbmuxdClient() as usb:
                    usb.delete_pair_record(self.udid)
                pair_record = self._get_pair_record() or self.pair_full()
                if not pair_record:
                    return False

        if resp.get('EnableSessionSSL'):
            if not pair_record:
                self.sslfile = get_home_path(self.cache_dir,f'{self.identifier}.pem')

            else:
                self.sslfile = write_home_file(
                    self.cache_dir,
                    f'{self.identifier}.pem',
                    pair_record['HostCertificate'] + b'\n' + pair_record['HostPrivateKey']
                )
            try:
                self.svc.ssl_start(self.sslfile, self.sslfile)
            except OSError:
                self.svc = PlistService(62078, self.udid, self.device, network=self.network)
                return False
        return True

    def pair_full(self):
        device_public_key = self.get_value(key='DevicePublicKey')
        wifi_address = self.get_value(key="WiFiAddress")
        with UsbmuxdClient() as usb:
            buid = usb.read_system_buid()
        if not device_public_key:
            log.error('Unable to retrieve DevicePublicKey')
            return
        log.debug('Creating host key & certificate')
        cert_pem, priv_key_pem, dev_cert_pem = make_certs_and_key(device_public_key)
        pair_record = {
            'DevicePublicKey': device_public_key,
            'DeviceCertificate': dev_cert_pem,
            'HostCertificate': cert_pem,
            'HostID': self.host_id,
            'RootCertificate': cert_pem,
            'SystemBUID': buid
        }

        pair = self.svc.plist_request({'Label': self.label,
                                       'Request': 'Pair',
                                       'PairRecord': pair_record,
                                       "ProtocolVersion": "2",
                                       "PairingOptions": {
                                           "ExtendedPairingErrors": True}
                                       })

        if pair and pair.get('Result') == 'Success' or 'EscrowBag' in pair:
            pair_record['HostPrivateKey'] = priv_key_pem
            pair_record['EscrowBag'] = pair.get('EscrowBag')
            pair_record['WiFiMACAddress'] = wifi_address
            with UsbmuxdClient() as usb:
                usb.save_pair_record(self.udid, pair_record, self.device_id)
            write_home_file(
                self.cache_dir,
                f'{self.identifier}.pem',
                pair_record['HostCertificate'] + b'\n' + pair_record['HostPrivateKey']
            )
            return pair_record
        elif pair and pair.get('Error') == 'PasswordProtected':
            self.svc.close()
            raise NotTrustedError
        else:
            log.error(pair.get('Error'))
            self.svc.close()
            raise PairingError

    def _plist_request(self, request: str, fields: Optional[Mapping[str, Any]] = None, label=True, **kwargs):
        req = {'Request': request, 'Label': self.label} if label else {'Request': request}
        if fields:
            req.update(fields)
        for k, v in kwargs.items():
            if v:
                req[k] = v
        return self.svc.plist_request(req)

    def get_value(self, domain='', key=None):
        if isinstance(key, str) and self.record and key in self.record:
            return self.record[key]
        resp = self._plist_request('GetValue', Domain=domain, Key=key)
        if resp:
            value = resp.get('Value')
            if hasattr(value, 'data'):
                return value.data
            return value
        return None

    def set_value(self, value, domain=None, key=None):
        resp = self._plist_request('SetValue', {'Value': value}, Domain=domain, Key=key)
        log.debug(f'set_value {resp}')
        if resp.get('Error'):
            log.error(f'set_value {resp}')
        return resp

    def remove_value(self, domain=None, key=None):
        resp = self._plist_request('RemoveValue', Domain=domain, Key=key)
        log.debug(f'remove_value {resp}')
        return resp

    def enable_wireless(self, enable, wireless_id=None, buddy_id=None):

        self.set_value(domain='com.apple.mobile.wireless_lockdown', key='EnableWifiConnections', value=enable)
        self.set_value(domain='com.apple.mobile.wireless_lockdown', key='EnableWifiDebugging', value=enable)
        if enable:
            if buddy_id is not None:
                # buddy_id = [str(uuid.uuid4())]
                self.set_value(domain='com.apple.mobile.wireless_lockdown', key='WirelessBuddyID', value=buddy_id)
            if wireless_id is None:
                wireless_id = ['']
            self.set_value(domain='com.apple.xcode.developerdomain', key='WirelessHosts', value=wireless_id)
        else:
            self.remove_value(domain='com.apple.xcode.developerdomain', key='WirelessHosts')

    def _start_service(self, name: str, escrow_bag=None) -> PlistService:
        if not self.paired:
            self._pair()
        elif not name:
            raise ValueError('Name must be a valid string')

        escrow_bag = self.record['EscrowBag'] if escrow_bag is True else escrow_bag
        resp = self._plist_request('StartService', Service=name, EscrowBag=escrow_bag)
        if not resp:
            raise StartServiceError(f'Unable to start service={name!r}')
        elif resp.get('Error'):
            if resp.get('Error') == 'PasswordProtected':
                raise StartServiceError(f'Unable to start service={name!r} - a password must be entered on the device')
            error = resp.get('Error')
            if self.ios_version >= LooseVersion('16.0'):
                log.info('try `pyidevice enable_developer_mode`')
            raise StartServiceError(f'Unable to start service={name!r} - {error}')
        log.debug(f'connect port: {resp.get("Port")}')
        plist_service = PlistService(
            resp.get('Port'), self.udid, ssl_file=self.sslfile if resp.get('EnableServiceSSL', False) else None,
            network=self.network)
        return plist_service

    @property
    def imagemounter(self):
        """
        start_service will call imagemounter, so here should call
        _unsafe_start_service instead
        """
        service = self._start_service("com.apple.mobile.mobile_image_mounter")
        return MobileImageMounter(service=service)

    def _urlretrieve(self, url, local_filename):
        """ download url to local """
        log.info("Download %s -> %s", url, local_filename)

        try:
            tmp_local_filename = local_filename + f".download-{int(time.time() * 1000)}"
            with requests.get(url, stream=True, timeout=DOWN_SUPPORT_TIMEOUT) as r:
                r.raise_for_status()
                with open(tmp_local_filename, 'wb') as f:
                    shutil.copyfileobj(r.raw, f, length=16 << 20)
                    f.flush()
                os.rename(tmp_local_filename, local_filename)
                size = os.path.getsize(local_filename)
                if size < DEVICE_SUPPORT_SIZE:
                    log.error("%r file download failed with abnormal file size", local_filename)
                    os.remove(local_filename)
                log.info("%r download successfully", local_filename)
        finally:
            if os.path.isfile(tmp_local_filename):
                os.remove(tmp_local_filename)

    @contextlib.contextmanager
    def _request_developer_image_dir(self):
        product_version = self.get_value('', "ProductVersion")
        log.info("ProductVersion: %s", product_version)
        major, minor = product_version.split(".")[:2]
        version = major + "." + minor

        mac_developer_dir = f"/Applications/Xcode.app/Contents/Developer/Platforms/iPhoneOS.platform/DeviceSupport/{version}"
        image_path = os.path.join(mac_developer_dir, "DeveloperDiskImage.dmg")
        signature_path = image_path + ".signature"
        if os.path.isfile(image_path) and os.path.isfile(signature_path):
            yield mac_developer_dir
        else:
            local_device_support = get_app_dir("device-support")
            image_zip_path = os.path.join(local_device_support, version + ".zip")
            if not os.path.isfile(image_zip_path):
                # origin_url = f"https://github.com/JinjunHan/iOSDeviceSupport/raw/master/DeviceSupport/{version}.zip"
                origin_url = f"https://github.com/JinjunHan/iOSDeviceSupport/tree/master/iOSDeviceSupport/{version}.zip"
                mirror_url = f"https://raw.githubusercontent.com/filsv/iPhoneOSDeviceSupport/master/{version}.zip"
                log.info("Download %s -> %s", origin_url, image_zip_path)
                try:
                    self._urlretrieve(mirror_url, image_zip_path)
                except requests.HTTPError:
                    log.debug("mirror download failed, change to original url")
                    # this might be slower
                    self._urlretrieve(origin_url, image_zip_path)

            with tempfile.TemporaryDirectory() as tmpdir:
                zf = zipfile.ZipFile(image_zip_path)
                zf.extractall(tmpdir)
                yield os.path.join(tmpdir, os.listdir(tmpdir)[0])

    def mount_developer_image(self):
        """
        Raises:
            MuxError
        """
        try:
            if self.imagemounter.is_developer_mounted():
                log.info("DeveloperImage already mounted")
                return
        except Exception:  # expect: DeviceLocked
            pass

        with self._request_developer_image_dir() as _dir:  # , signature_path:
            image_path = os.path.join(_dir, "DeveloperDiskImage.dmg")
            signature_path = image_path + ".signature"
            self.imagemounter.mount(image_path, signature_path)
            log.info("DeveloperImage mounted successfully")

    def start_service(self, name: str, escrow_bag=None) -> PlistService:
        try:
            return self._start_service(name, escrow_bag)
        except StartServiceError:
            self.mount_developer_image()
            time.sleep(.5)
            return self._start_service(name, escrow_bag)

    def stop_session(self):
        if self.session_id and self.svc:
            resp = self._plist_request('StopSession', SessionID=self.session_id)
            self.session_id = None
            if not resp or resp.get('Result') != 'Success':
                raise CannotStopSessionError(resp)
            return resp

    def enter_recovery(self):
        log.debug(self.svc.plist_request({'Request': 'EnterRecovery'}))


def get_home_path(foldername: str, filename: str) -> Path:
    path = Path('~').expanduser().joinpath(foldername)
    if not path.exists():
        path.mkdir(parents=True)
    return path.joinpath(filename)


def read_home_file(foldername: str, filename: str) -> Optional[bytes]:
    path = get_home_path(foldername, filename)
    if not path.exists():
        return None
    with path.open('rb') as f:
        return f.read()


def write_home_file(foldername: str, filename: str, data: bytes) -> str:
    path = get_home_path(foldername, filename)
    log.debug(f'save path :{path}')
    with path.open('wb') as f:
        f.write(data)
    return path.as_posix()


def _get_lockdown_dir():
    if sys.platform == 'win32':
        return Path(os.environ['ALLUSERSPROFILE'] + '/Apple/Lockdown/')
    else:
        return Path('/var/db/lockdown/')



import time
"""
获取单应用的网络信息
"""
from ios_device.servers.Instrument import InstrumentServer


def netstat(rpc, pid):
    rpc._start()
    channel = "com.apple.xcode.debug-gauge-data-providers.NetworkStatistics"
    attr = {}
    print("start", rpc.call(channel, "startSamplingForPIDs:", {pid}).selector)
    ret = rpc.call(channel, "sampleAttributes:forPIDs:", attr, {pid})
    print(ret.selector)
    rpc.stop()


if __name__ == '__main__':
    rpc = InstrumentServer().init()
    netstat(rpc, 261)
    rpc.stop()



import json
import os
import sys


sys.path.append(os.getcwd())
import time
from _ctypes import Structure
from ctypes import c_byte, c_uint16, c_uint32
from ios_device.servers.Instrument import  InstrumentServer
from ios_device.util import logging

log = logging.getLogger(__name__)


def networking(rpc):
    headers = {
        0: ['InterfaceIndex', "Name"],
        1: ['LocalAddress', 'RemoteAddress', 'InterfaceIndex', 'Pid', 'RecvBufferSize', 'RecvBufferUsed',
            'SerialNumber', 'Kind'],
        2: ['RxPackets', 'RxBytes', 'TxPackets', 'TxBytes', 'RxDups', 'RxOOO', 'TxRetx', 'MinRTT', 'AvgRTT',
            'ConnectionSerial']
    }
    msg_type = {
        0: "interface-detection",
        1: "connection-detected",
        2: "connection-update",
    }

    def on_callback_message(res):
        from socket import inet_ntoa, htons, inet_ntop, AF_INET6
        class SockAddr4(Structure):
            _fields_ = [
                ('len', c_byte),
                ('family', c_byte),
                ('port', c_uint16),
                ('addr', c_byte * 4),
                ('zero', c_byte * 8)
            ]

            def __str__(self):
                return f"{inet_ntoa(self.addr)}:{htons(self.port)}"

        class SockAddr6(Structure):
            _fields_ = [
                ('len', c_byte),
                ('family', c_byte),
                ('port', c_uint16),
                ('flowinfo', c_uint32),
                ('addr', c_byte * 16),
                ('scopeid', c_uint32)
            ]

            def __str__(self):
                return f"[{inet_ntop(AF_INET6, self.addr)}]:{htons(self.port)}"

        data = res.selector
        if data[0] == 1:
            if len(data[1][0]) == 16:
                data[1][0] = str(SockAddr4.from_buffer_copy(data[1][0]))
                data[1][1] = str(SockAddr4.from_buffer_copy(data[1][1]))
            elif len(data[1][0]) == 28:
                data[1][0] = str(SockAddr6.from_buffer_copy(data[1][0]))
                data[1][1] = str(SockAddr6.from_buffer_copy(data[1][1]))

        print(msg_type[data[0]] + json.dumps(dict(zip(headers[data[0]], data[1]))))
        # print("[data]", res.selector)
    rpc.register_channel_callback("com.apple.instruments.server.services.networking", on_callback_message)
    var = rpc.call("com.apple.instruments.server.services.networking", "replayLastRecordedSession").selector
    log.debug(f"replay {var}")
    var = rpc.call("com.apple.instruments.server.services.networking", "startMonitoring").selector
    log.debug(f"start {var}")
    time.sleep(10)
    var = rpc.call("com.apple.instruments.server.services.networking", "stopMonitoring").selector
    log.debug(f"stopMonitoring {var}")
    rpc.stop()


if __name__ == '__main__':
    rpc = InstrumentServer().init()
    networking(rpc)
    rpc.stop()



# Referenced from
# https://opensource.apple.com/tarballs/xnu-7195.60.75/bsd/kern/kdebug.c
# https://gitee.com/mirrors/darwin-xnu/blob/main/bsd/kern/kdebug.c  kdebug.h


import enum
import io
import json
import struct

from construct import Struct, Const, Padding, Int32ul, Int64ul, Array, GreedyRange, Byte, FixedSized, \
    CString, GreedyBytes, this, Adapter, Int16ul, Switch, StreamError

from ios_device.util import plistlib
from ios_device.util.kc_data import kc_data_parse

# from pykdebugparser.kd_buf_parser import BplistAdapter

KDBG_CLASS_MASK = 0xff000000
KDBG_CLASS_OFFSET = 24
KDBG_CLASS_MAX = 0xff

KDBG_SUBCLASS_MASK = 0x00ff0000
KDBG_SUBCLASS_OFFSET = 16
KDBG_SUBCLASS_MAX = 0xff

# ## /* class and subclass mask */
KDBG_CSC_MASK = 0xffff0000
KDBG_CSC_OFFSET = KDBG_SUBCLASS_OFFSET
KDBG_CSC_MAX = 0xffff

KDBG_CODE_MASK = 0x0000fffc
KDBG_CODE_OFFSET = 2
KDBG_CODE_MAX = 0x3fff

KDBG_EVENTID_MASK = 0xfffffffc
KDBG_FUNC_MASK = 0x00000003
VERSION2_FLAG = b'\x00\x02\xaa\x55'
VERSION3_FLAG = b'\x00\x03\xaa\x55'
KD_BUF_FORMAT = '<Q32sQLLQ'


def kdbg_extract_class(Debugid):
    return (Debugid & KDBG_CLASS_MASK) >> KDBG_CLASS_OFFSET


def kdbg_extract_subclass(Debugid):
    return (Debugid & KDBG_SUBCLASS_MASK) >> KDBG_SUBCLASS_OFFSET


def kdbg_extract_csc(Debugid):
    return (Debugid & KDBG_CSC_MASK) >> KDBG_CSC_OFFSET


def kdbg_extract_code(Debugid):
    return (Debugid & KDBG_CODE_MASK) >> KDBG_CODE_OFFSET


class DgbFuncQual(enum.Enum):
    """
    ## /* function qualifiers  */
        DBG_FUNC_START 1U
        DBG_FUNC_END   2U
        DBG_FUNC_NONE = 0U
    Event's role in the trace.
    """
    DBG_FUNC_NONE = 0
    DBG_FUNC_START = 1
    DBG_FUNC_END = 2
    DBG_FUNC_ALL = 3


class DebugClasses(enum.Enum):
    DBG_MACH = 1
    DBG_NETWORK = 2
    DBG_FSYSTEM = 3
    DBG_BSD = 4
    DBG_IOKIT = 5
    DBG_DRIVERS = 6
    DBG_TRACE = 7
    DBG_DLIL = 8
    DBG_PTHREAD = 9
    DBG_CORESTORAGE = 10
    DBG_CG = 11
    DBG_MONOTONIC = 12
    DBG_MISC = 20
    DBG_SECURITY = 30
    DBG_DYLD = 31
    DBG_QT = 32
    DBG_APPS = 33
    DBG_LAUNCHD = 34
    DBG_SILICON = 35
    DBG_PERF = 37
    DBG_IMPORTANCE = 38
    DBG_BANK = 40
    DBG_XPC = 41
    DBG_ATM = 42
    DBG_ARIADNE = 43
    DBG_DAEMON = 44
    DBG_ENERGYTRACE = 45
    DBG_DISPATCH = 46
    DBG_IMG = 49
    DBG_UMALLOC = 51
    DBG_TURNSTILE = 53
    DBG_MIG = 255


class DBG_MACH(enum.Enum):
    DBG_MACH_EXCP_KTRAP_x86 = 0x02  ## /* Kernel Traps on x86 */
    DBG_MACH_EXCP_DFLT = 0x03  ## /* deprecated name */
    DBG_MACH_EXCP_SYNC_ARM = 0x03  ## /* arm/arm64 synchronous exception */
    DBG_MACH_EXCP_IFLT = 0x04  ## /* deprecated name */
    DBG_MACH_EXCP_SERR_ARM = 0x04  ## /* arm/arm64 SError (async) exception */
    DBG_MACH_EXCP_INTR = 0x05  ## /* Interrupts */
    DBG_MACH_EXCP_ALNG = 0x06  ## /* Alignment Exception */
    DBG_MACH_EXCP_UTRAP_x86 = 0x07  ## /* User Traps on x86 */
    DBG_MACH_EXCP_FP = 0x08  ## /* FP Unavail */
    DBG_MACH_EXCP_DECI = 0x09  ## /* Decrementer Interrupt */
    DBG_MACH_CHUD = 0x0A  ## /* deprecated name */
    DBG_MACH_SIGNPOST = 0x0A  ## /* kernel signposts */
    DBG_MACH_EXCP_SC = 0x0C  ## /* System Calls */
    DBG_MACH_EXCP_TRACE = 0x0D  ## /* Trace exception */
    DBG_MACH_EXCP_EMUL = 0x0E  ## /* Instruction emulated */
    DBG_MACH_IHDLR = 0x10  ## /* Interrupt Handlers */
    DBG_MACH_IPC = 0x20  ## /* Inter Process Comm */
    DBG_MACH_RESOURCE = 0x25  ## /* tracing limits, etc */
    DBG_MACH_VM = 0x30  ## /* Virtual Memory */
    DBG_MACH_LEAKS = 0x31  ## /* alloc/free */
    DBG_MACH_WORKINGSET = 0x32  ## /* private subclass for working set related debugging */
    DBG_MACH_SCHED = 0x40  ## /* Scheduler */
    DBG_MACH_MSGID_INVALID = 0x50  ## /* Messages - invalid */
    DBG_MACH_LOCKS = 0x60  ## /* new lock APIs */
    DBG_MACH_PMAP = 0x70  ## /* pmap */
    DBG_MACH_CLOCK = 0x80  ## /* clock */
    DBG_MACH_MP = 0x90  ## /* MP related */
    DBG_MACH_VM_PRESSURE = 0xA0  ## /* Memory Pressure Events */
    DBG_MACH_STACKSHOT = 0xA1  ## /* Stackshot/Microstackshot subsystem */
    DBG_MACH_SFI = 0xA2  ## /* Selective Forced Idle (SFI) */
    DBG_MACH_ENERGY_PERF = 0xA3  ## /* Energy/performance resource stats */
    DBG_MACH_SYSDIAGNOSE = 0xA4  ## /* sysdiagnose */
    DBG_MACH_ZALLOC = 0xA5  ## /* Zone allocator */
    DBG_MACH_THREAD_GROUP = 0xA6  ## /* Thread groups */
    DBG_MACH_COALITION = 0xA7  ## /* Coalitions */
    DBG_MACH_SHAREDREGION = 0xA8  ## /* Shared region */
    DBG_MACH_SCHED_CLUTCH = 0xA9  ## /* Clutch scheduler */
    DBG_MACH_IO = 0xAA  ## /* I/O */
    DBG_MACH_WORKGROUP = 0xAB  ## /* Workgroup subsystem */


class DBG_MACH_IO(enum.Enum):
    DBC_MACH_IO_MMIO_READ = 0x1
    DBC_MACH_IO_MMIO_WRITE = 0x2
    DBC_MACH_IO_PHYS_READ = 0x3
    DBC_MACH_IO_PHYS_WRITE = 0x4
    DBC_MACH_IO_PORTIO_READ = 0x5
    DBC_MACH_IO_PORTIO_WRITE = 0x6


class DBG_MACH_EXCP_INTR(enum.Enum):
    DBG_INTR_TYPE_UNKNOWN = 0x0  ## /* default/unknown interrupt */
    DBG_INTR_TYPE_IPI = 0x1  ## /* interprocessor interrupt */
    DBG_INTR_TYPE_TIMER = 0x2  ## /* timer interrupt */
    DBG_INTR_TYPE_OTHER = 0x3  ## /* other (usually external) interrupt */
    DBG_INTR_TYPE_PMI = 0x4  ## /* performance monitor interrupt */


class DBG_MACH_SCHED(enum.Enum):
    MACH_SCHED = 0x0  # /* Scheduler */
    MACH_STACK_ATTACH = 0x1  # /* stack_attach() */
    MACH_STACK_HANDOFF = 0x2  # /* stack_handoff() */
    MACH_CALL_CONT = 0x3  # /* call_continuation() */
    MACH_CALLOUT = 0x4  # /* callouts */
    MACH_STACK_DETACH = 0x5
    MACH_MAKE_RUNNABLE = 0x6  # /* make thread runnable */
    MACH_PROMOTE = 0x7  # /* promoted due to resource (replaced by MACH_PROMOTED) */
    MACH_DEMOTE = 0x8  # /* promotion undone (replaced by MACH_UNPROMOTED) */
    MACH_IDLE = 0x9  # /* processor idling */
    MACH_STACK_DEPTH = 0xa  # /* stack depth at switch */
    MACH_MOVED = 0xb  # /* did not use original scheduling decision */
    MACH_PSET_LOAD_AVERAGE = 0xc
    MACH_AMP_DEBUG = 0xd
    MACH_FAILSAFE = 0xe  # /* tripped fixed-pri/RT failsafe */
    MACH_BLOCK = 0xf  # /* thread block */
    MACH_WAIT = 0x10  # /* thread wait assertion */
    MACH_GET_URGENCY = 0x14  # /* Urgency queried by platform */
    MACH_URGENCY = 0x15  # /* Urgency (RT/BG/NORMAL) communicated
    MACH_REDISPATCH = 0x16  # /* "next thread" thread redispatched */
    MACH_REMOTE_AST = 0x17  # /* AST signal issued to remote processor */
    MACH_SCHED_CHOOSE_PROCESSOR = 0x18  # /* Result of choose_processor */
    MACH_DEEP_IDLE = 0x19  # /* deep idle on master processor */
    MACH_SCHED_DECAY_PRIORITY = 0x1a  # /*was MACH_SCHED_DECAY_PRIORITY */
    MACH_CPU_THROTTLE_DISABLE = 0x1b  # /* Global CPU Throttle Disable */
    MACH_RW_PROMOTE = 0x1c  # /* promoted due to RW lock promotion */
    MACH_RW_DEMOTE = 0x1d  # /* promotion due to RW lock undone */
    MACH_SCHED_MAINTENANCE = 0x1f  # /* periodic maintenance thread */
    MACH_DISPATCH = 0x20  # /* context switch completed */
    MACH_QUANTUM_HANDOFF = 0x21  # /* quantum handoff occurred */
    MACH_MULTIQ_DEQUEUE = 0x22  # /* Result of multiq dequeue */
    MACH_SCHED_THREAD_SWITCH = 0x23  # /* attempt direct context switch to hinted thread */
    MACH_SCHED_SMT_BALANCE = 0x24  # /* SMT load balancing ASTs */
    MACH_REMOTE_DEFERRED_AST = 0x25  # /* Deferred AST started against remote processor */
    MACH_REMOTE_CANCEL_AST = 0x26  # /* Canceled deferred AST for remote processor */
    MACH_SCHED_CHANGE_PRIORITY = 0x27  # /* thread sched priority changed */
    MACH_SCHED_UPDATE_REC_CORES = 0x28  # /* Change to recommended processor bitmask */
    MACH_STACK_WAIT = 0x29  # /* Thread could not be switched-to because of kernel stack shortage */
    MACH_THREAD_BIND = 0x2a  # /* Thread was bound (or unbound) to a processor */
    MACH_WAITQ_PROMOTE = 0x2b  # /* Thread promoted by waitq boost */
    MACH_WAITQ_DEMOTE = 0x2c  # /* Thread demoted from waitq boost */
    MACH_SCHED_LOAD = 0x2d  # /* load update */
    MACH_REC_CORES_FAILSAFE = 0x2e  # /* recommended processor failsafe kicked in */
    MACH_SCHED_QUANTUM_EXPIRED = 0x2f  # /* thread quantum expired */
    MACH_EXEC_PROMOTE = 0x30  # /* Thread promoted by exec boost */
    MACH_EXEC_DEMOTE = 0x31  # /* Thread demoted from exec boost */
    MACH_AMP_SIGNAL_SPILL = 0x32  # /* AMP spill signal sent to cpuid */
    MACH_AMP_STEAL = 0x33  # /* AMP thread stolen or spilled */
    MACH_SCHED_LOAD_EFFECTIVE = 0x34  # /* Effective scheduler load */
    MACH_PROMOTED = 0x35  # /* was: thread promoted due to mutex priority promotion */
    MACH_UNPROMOTED = 0x36  # /* was: thread unpromoted due to mutex priority promotion */
    MACH_PROMOTED_UPDATE = 0x37  # /* was: thread already promoted, but promotion priority changed */
    MACH_QUIESCENT_COUNTER = 0x38  # /* quiescent counter tick */
    MACH_TURNSTILE_USER_CHANGE = 0x39  # /* base priority change because of turnstile */
    MACH_AMP_RECOMMENDATION_CHANGE = 0x3a  # /* Thread group recommendation change */
    MACH_AMP_PERFCTL_POLICY_CHANGE = 0x3b  # /* AMP policy for perfctl cluster recommendation */
    MACH_TURNSTILE_KERNEL_CHANGE = 0x40  # /* sched priority change because of turnstile */
    MACH_SCHED_WI_AUTO_JOIN = 0x41  # /* work interval auto join events */
    MACH_SCHED_WI_DEFERRED_FINISH = 0x42  # /* work interval pending finish events for auto-join thread groups */
    MACH_PSET_AVG_EXEC_TIME = 0x50


class DBG_MACH_SCHED_CLUTCH(enum.Enum):
    MACH_SCHED_CLUTCH_ROOT_BUCKET_STATE = 0x0  # /* __unused */
    MACH_SCHED_CLUTCH_TG_BUCKET_STATE = 0x1  # /* __unused */
    MACH_SCHED_CLUTCH_THREAD_SELECT = 0x2  # /* Thread selection events for Clutch scheduler */
    MACH_SCHED_CLUTCH_THREAD_STATE = 0x3  # /* __unused */
    MACH_SCHED_CLUTCH_TG_BUCKET_PRI = 0x4  # /* Clutch bucket priority update event */
    MACH_SCHED_EDGE_CLUSTER_OVERLOAD = 0x5  # /* Cluster experienced overload; migrating threads to other clusters */
    MACH_SCHED_EDGE_STEAL = 0x6  # /* Per-cluster avg. thread execution time */
    MACH_SCHED_EDGE_REBAL_RUNNABLE = 0x7  # /* Rebalance runnable threads on a foreign cluster */
    MACH_SCHED_EDGE_REBAL_RUNNING = 0x8  # /* Rebalance running threads on a foreign cluster */
    MACH_SCHED_EDGE_SHOULD_YIELD = 0x9  # /* Edge decisions for thread yield */
    MACH_SCHED_CLUTCH_THR_COUNT = 0xa  # /* Clutch scheduler runnable thread counts */
    MACH_SCHED_EDGE_LOAD_AVG = 0xb  # /* Per-cluster load average */


class DBG_MACH_WORKGROUP(enum.Enum):
    WORKGROUP_INTERVAL_CREATE = 0x0  # /* work interval creation */
    WORKGROUP_INTERVAL_DESTROY = 0x1  # /* work interval destruction */
    WORKGROUP_INTERVAL_CHANGE = 0x2  # /* thread work interval change */
    WORKGROUP_INTERVAL_START = 0x3  # /* work interval start call */
    WORKGROUP_INTERVAL_UPDATE = 0x4  # /* work interval update call */
    WORKGROUP_INTERVAL_FINISH = 0x5  # /* work interval finish call */


class DBG_MACH_IPC(enum.Enum):
    MACH_TASK_SUSPEND = 0x0  # /* Suspended a task */
    MACH_TASK_RESUME = 0x1  # /* Resumed a task */
    MACH_THREAD_SET_VOUCHER = 0x2
    MACH_IPC_MSG_SEND = 0x3  # /* mach msg send, uniq msg info */
    MACH_IPC_MSG_RECV = 0x4  # /* mach_msg receive */
    MACH_IPC_MSG_RECV_VOUCHER_REFUSED = 0x5  # /* mach_msg receive, voucher refused */
    MACH_IPC_KMSG_FREE = 0x6  # /* kernel free of kmsg data */
    MACH_IPC_VOUCHER_CREATE = 0x7  # /* Voucher added to global voucher hashtable */
    MACH_IPC_VOUCHER_CREATE_ATTR_DATA = 0x8  # /* Attr data for newly created voucher */
    MACH_IPC_VOUCHER_DESTROY = 0x9  # /* Voucher removed from global voucher hashtable */
    MACH_IPC_KMSG_INFO = 0xa  # /* Send/Receive info for a kmsg */
    MACH_IPC_KMSG_LINK = 0xb  # /* link a kernel kmsg pointer to user mach_msg_header_t */
    MACH_IPC_PORT_ENTRY_MODIFY = 0xc  # /* A port space gained or lost a port right (reference) */
    MACH_IPC_DESTROY_GUARDED_DESC = 0xd  # /* Unable to receive a guarded descriptor */


class DBG_MACH_THREAD_GROUP(enum.Enum):
    MACH_THREAD_GROUP_NEW = 0x0
    MACH_THREAD_GROUP_FREE = 0x1
    MACH_THREAD_GROUP_SET = 0x2
    MACH_THREAD_GROUP_NAME = 0x3
    MACH_THREAD_GROUP_NAME_FREE = 0x4
    MACH_THREAD_GROUP_FLAGS = 0x5
    MACH_THREAD_GROUP_BLOCK = 0x6


class DBG_MACH_COALITION(enum.Enum):
    MACH_COALITION_NEW = 0x0
    MACH_COALITION_FREE = 0x1
    MACH_COALITION_ADOPT = 0x2
    MACH_COALITION_REMOVE = 0x3
    MACH_COALITION_THREAD_GROUP_SET = 0x4


class DBG_MACH_PMAP(enum.Enum):
    MACH_COALITION_NEW = 0x0
    MACH_COALITION_FREE = 0x1
    MACH_COALITION_ADOPT = 0x2
    MACH_COALITION_REMOVE = 0x3
    MACH_COALITION_THREAD_GROUP_SET = 0x4


class DBG_MACH_CLOCK(enum.Enum):
    MACH_EPOCH_CHANGE = 0x0  # /* wake epoch change */
    MACH_BRIDGE_RCV_TS = 0x1  # /* receive timestamp pair from interrupt handler */
    MACH_BRIDGE_REMOTE_TIME = 0x2  # /* calculate remote timestamp */
    MACH_BRIDGE_RESET_TS = 0x3  # /* reset timestamp conversion parameters */
    MACH_BRIDGE_TS_PARAMS = 0x4  # /* recompute timestamp conversion parameters */
    MACH_BRIDGE_SKIP_TS = 0x5  # /* skip timestamp */
    MACH_BRIDGE_TS_MISMATCH = 0x6  # /* mismatch between predicted and received remote timestamp */
    MACH_BRIDGE_OBSV_RATE = 0x7  # /* out of range observed rates */


class DBG_MACH_STACKSHOT(enum.Enum):
    MICROSTACKSHOT_RECORD = 0x0
    MICROSTACKSHOT_GATHER = 0x1


class DBG_MACH_SYSDIAGNOSE(enum.Enum):
    SYSDIAGNOSE_NOTIFY_USER = 0x0
    SYSDIAGNOSE_FULL = 0x1
    SYSDIAGNOSE_STACKSHOT = 0x2
    SYSDIAGNOSE_TAILSPIN = 0x3


class DBG_MACH_SFI(enum.Enum):
    SFI_SET_WINDOW = 0x0
    SFI_CANCEL_WINDOW = 0x1
    SFI_SET_CLASS_OFFTIME = 0x2
    SFI_CANCEL_CLASS_OFFTIME = 0x3
    SFI_THREAD_DEFER = 0x4
    SFI_OFF_TIMER = 0x5
    SFI_ON_TIMER = 0x6
    SFI_WAIT_CANCELED = 0x7
    SFI_PID_SET_MANAGED = 0x8
    SFI_PID_CLEAR_MANAGED = 0x9
    SFI_GLOBAL_DEFER = 0xa


class DBG_MACH_ZALLOC(enum.Enum):
    ZALLOC_ZCRAM = 0x0


class DBG_MACH_RESOURCE(enum.Enum):
    RMON_ENABLE_CPUUSAGE_MONITOR = 0x001
    RMON_CPUUSAGE_VIOLATED = 0x002
    RMON_CPUUSAGE_SUSPENDED = 0x003
    RMON_CPUUSAGE_VIOLATED_K32A = 0x004
    RMON_CPUUSAGE_VIOLATED_K32B = 0x005
    RMON_CPUUSAGE_RESUMED = 0x006
    RMON_DISABLE_CPUUSAGE_MONITOR = 0x00f

    RMON_ENABLE_CPUWAKES_MONITOR = 0x011
    RMON_CPUWAKES_VIOLATED = 0x012
    RMON_CPUWAKES_VIOLATED_K32A = 0x014
    RMON_CPUWAKES_VIOLATED_K32B = 0x015
    RMON_DISABLE_CPUWAKES_MONITOR = 0x01f

    RMON_ENABLE_IO_MONITOR = 0x021
    RMON_LOGWRITES_VIOLATED = 0x022
    RMON_PHYSWRITES_VIOLATED = 0x023
    RMON_LOGWRITES_VIOLATED_K32A = 0x024
    RMON_LOGWRITES_VIOLATED_K32B = 0x025
    RMON_DISABLE_IO_MONITOR = 0x02f


class DBG_NETWORK(enum.Enum):
    DBG_NETIP = 1  # /* Internet Protocol */
    DBG_NETARP = 2  # /* Address Resolution Protocol */
    DBG_NETUDP = 3  # /* User Datagram Protocol */
    DBG_NETTCP = 4  # /* Transmission Control Protocol */
    DBG_NETICMP = 5  # /* Internet Control Message Protocol */
    DBG_NETIGMP = 6  # /* Internet Group Management Protocol */
    DBG_NETRIP = 7  # /* Routing Information Protocol */
    DBG_NETOSPF = 8  # /* Open Shortest Path First */
    DBG_NETISIS = 9  # /* Intermediate System to Intermediate System */
    DBG_NETSNMP = 10  # /* Simple Network Management Protocol */
    DBG_NETSOCK = 11  # /* Socket Layer */
    DBG_NETAARP = 100  # /* Apple ARP */
    DBG_NETDDP = 101  # /* Datagram Delivery Protocol */
    DBG_NETNBP = 102  # /* Name Binding Protocol */
    DBG_NETZIP = 103  # /* Zone Information Protocol */
    DBG_NETADSP = 104  # /* Name Binding Protocol */
    DBG_NETATP = 105  # /* Apple Transaction Protocol */
    DBG_NETASP = 106  # /* Apple Session Protocol */
    DBG_NETAFP = 107  # /* Apple Filing Protocol */
    DBG_NETRTMP = 108  # /* Routing Table Maintenance Protocol */
    DBG_NETAURP = 109  # /* Apple Update Routing Protocol */

    DBG_NETIPSEC = 128  # /* IPsec Protocol  */
    DBG_NETVMNET = 129  # /* VMNet */


class DBG_IOKIT(enum.Enum):
    DBG_IOINTC = 0  # /* Interrupt controller */
    DBG_IOWORKLOOP = 1  # /* Work from work loop */
    DBG_IOINTES = 2  # /* Interrupt event source */
    DBG_IOCLKES = 3  # /* Clock event source */
    DBG_IOCMDQ = 4  # /* Command queue latencies */
    DBG_IOMCURS = 5  # /* Memory Cursor */
    DBG_IOMDESC = 6  # /* Memory Descriptors */
    DBG_IOPOWER = 7  # /* Power Managerment */
    DBG_IOSERVICE = 8  # /* Matching etc. */
    DBG_IOREGISTRY = 9  # /* Registry */

    # **** 9-32 reserved for internal IOKit usage ****

    DBG_IOSTORAGE = 32  # /* Storage layers */
    DBG_IONETWORK = 33  # /* Network layers */
    DBG_IOKEYBOARD = 34  # /* Keyboard */
    DBG_IOHID = 35  # /* HID Devices */
    DBG_IOAUDIO = 36  # /* Audio */
    DBG_IOSERIAL = 37  # /* Serial */
    DBG_IOTTY = 38  # /* TTY layers */
    DBG_IOSAM = 39  # /* SCSI Architecture Model layers */
    DBG_IOPARALLELATA = 40  # /* Parallel ATA */
    DBG_IOPARALLELSCSI = 41  # /* Parallel SCSI */
    DBG_IOSATA = 42  # /* Serial-ATA */
    DBG_IOSAS = 43  # /* SAS */
    DBG_IOFIBRECHANNEL = 44  # /* FiberChannel */
    DBG_IOUSB = 45  # /* USB */
    DBG_IOBLUETOOTH = 46  # /* Bluetooth */
    DBG_IOFIREWIRE = 47  # /* FireWire */
    DBG_IOINFINIBAND = 48  # /* Infiniband */
    DBG_IOCPUPM = 49  # /* CPU Power Management */
    DBG_IOGRAPHICS = 50  # /* Graphics */
    DBG_HIBERNATE = 51  # /* hibernation related events */
    DBG_IOTHUNDERBOLT = 52  # /* Thunderbolt */
    DBG_BOOTER = 53  # /* booter related events */
    DBG_IOAUDIO2 = 54  # /* Audio (extended) */

    DBG_IOSURFACEPA = 64  # /* IOSurface page mappings */
    DBG_IOMDPA = 65  # /* IOMemoryDescriptor page mappings */
    DBG_IODARTPA = 66  # /* DART page mappings */


class DBG_DRIVERS(enum.Enum):
    DBG_DRVSTORAGE = 1  # /* Storage layers */
    DBG_DRVNETWORK = 2  # /* Network layers */
    DBG_DRVKEYBOARD = 3  # /* Keyboard */
    DBG_DRVHID = 4  # /* HID Devices */
    DBG_DRVAUDIO = 5  # /* Audio */
    DBG_DRVSERIAL = 7  # /* Serial */
    DBG_DRVSAM = 8  # /* SCSI Architecture Model layers */
    DBG_DRVPARALLELATA = 9  # /* Parallel ATA */
    DBG_DRVPARALLELSCSI = 10  # /* Parallel SCSI */
    DBG_DRVSATA = 11  # /* Serial ATA */
    DBG_DRVSAS = 12  # /* SAS */
    DBG_DRVFIBRECHANNEL = 13  # /* FiberChannel */
    DBG_DRVUSB = 14  # /* USB */
    DBG_DRVBLUETOOTH = 15  # /* Bluetooth */
    DBG_DRVFIREWIRE = 16  # /* FireWire */
    DBG_DRVINFINIBAND = 17  # /* Infiniband */
    DBG_DRVGRAPHICS = 18  # /* Graphics */
    DBG_DRVSD = 19  # /* Secure Digital */
    DBG_DRVNAND = 20  # /* NAND drivers and layers */
    DBG_SSD = 21  # /* SSD */
    DBG_DRVSPI = 22  # /* SPI */
    DBG_DRVWLAN_802_11 = 23  # /* WLAN 802.11 */
    DBG_DRVSSM = 24  # /* System State Manager(AppleSSM) */
    DBG_DRVSMC = 25  # /* System Management Controller */
    DBG_DRVMACEFIMANAGER = 26  # /* Mac EFI Manager */
    DBG_DRVANE = 27  # /* ANE */
    DBG_DRVETHERNET = 28  # /* Ethernet */
    DBG_DRVMCC = 29  # /* Memory Cache Controller */
    DBG_DRVACCESSORY = 30  # /* Accessories */


class DBG_DLIL(enum.Enum):
    DBG_DLIL_STATIC = 1  # /* Static DLIL code */
    DBG_DLIL_PR_MOD = 2  # /* DLIL Protocol Module */
    DBG_DLIL_IF_MOD = 3  # /* DLIL Interface Module */
    DBG_DLIL_PR_FLT = 4  # /* DLIL Protocol Filter */
    DBG_DLIL_IF_FLT = 5  # /* DLIL Interface FIlter */


class DBG_FSYSTEM(enum.Enum):
    DBG_FSRW = 0x1  # /* reads and writes to the filesystem */
    DBG_DKRW = 0x2  # /* reads and writes to the disk */
    DBG_FSVN = 0x3  # /* vnode operations (inc. locking/unlocking) */
    DBG_FSLOOOKUP = 0x4  # /* namei and other lookup-related operations */
    DBG_JOURNAL = 0x5  # /* journaling operations */
    DBG_IOCTL = 0x6  # /* ioctl to the disk */
    DBG_BOOTCACHE = 0x7  # /* bootcache operations */
    DBG_HFS = 0x8  # /* HFS-specific events; see the hfs project */
    DBG_APFS = 0x9  # /* APFS-specific events; see the apfs project */
    DBG_SMB = 0xA  # /* SMB-specific events; see the smb project */
    DBG_MOUNT = 0xB  # /* Mounting/unmounting operations */
    DBG_EXFAT = 0xE  # /* ExFAT-specific events; see the exfat project */
    DBG_MSDOS = 0xF  # /* FAT-specific events; see the msdosfs project */
    DBG_ACFS = 0x10  # /* Xsan-specific events; see the XsanFS project */
    DBG_THROTTLE = 0x11  # /* I/O Throttling events */
    DBG_DECMP = 0x12  # /* Decmpfs-specific events */
    DBG_VFS = 0x13  # /* VFS layer events */
    DBG_LIVEFS = 0x14  # /* LiveFS events; see the UserFS project */
    DBG_CONTENT_PROT = 0xCF  # /* Content Protection Events: see bsd/sys/cprotect.h */


class DBG_BSD(enum.Enum):
    DBG_BSD_PROC = 0x01  # /* process/signals related */
    DBG_BSD_MEMSTAT = 0x02  # /* memorystatus / jetsam operations */
    DBG_BSD_KEVENT = 0x03  # /* kqueue / kevent related */
    DBG_BSD_EXCP_SC = 0x0C  # /* System Calls */
    DBG_BSD_AIO = 0x0D  # /* aio (POSIX async IO) */
    DBG_BSD_SC_EXTENDED_INFO = 0x0E  # /* System Calls, extended info */
    DBG_BSD_SC_EXTENDED_INFO2 = 0x0F  # /* System Calls, extended info */
    DBG_BSD_KDEBUG_TEST = 0xFF  # /* for testing kdebug */


class DBG_BSD_PROC(enum.Enum):
    BSD_PROC_EXIT = 1  # /* process exit */
    BSD_PROC_FRCEXIT = 2  # /* Kernel force termination */
    BSD_PROC_EXEC = 3  # /* process spawn / exec */
    BSD_PROC_EXITREASON_CREATE = 4  # /* exit reason creation */
    BSD_PROC_EXITREASON_COMMIT = 5  # /* exit reason commited to a proc */


class DBG_BSD_MEMSTAT(enum.Enum):
    BSD_MEMSTAT_SCAN = 1  # /* memorystatus thread awake */
    BSD_MEMSTAT_JETSAM = 2  # /* LRU jetsam */
    BSD_MEMSTAT_JETSAM_HIWAT = 3  # /* highwater jetsam */
    BSD_MEMSTAT_FREEZE = 4  # /* freeze process */
    BSD_MEMSTAT_FREEZE_SCAN = 5  # /* select a process to freeze and freeze it */
    BSD_MEMSTAT_UPDATE = 6  # /* priority update */
    BSD_MEMSTAT_IDLE_DEMOTE = 7  # /* idle demotion fired */
    BSD_MEMSTAT_CLEAR_ERRORS = 8  # /* reset termination error state */
    BSD_MEMSTAT_DIRTY_TRACK = 9  # /* track the process state */
    BSD_MEMSTAT_DIRTY_SET = 10  # /* set the process state */
    BSD_MEMSTAT_DIRTY_CLEAR = 11  # /* clear the process state */
    BSD_MEMSTAT_GRP_SET_PROP = 12  # /* set group properties */
    BSD_MEMSTAT_DO_KILL = 13  # /* memorystatus kills */
    BSD_MEMSTAT_CHANGE_PRIORITY = 14  # /* priority changed */
    BSD_MEMSTAT_FAST_JETSAM = 15  # /* Aggressive jetsam ("clear-the-deck") */
    BSD_MEMSTAT_COMPACTOR_RUN = 16  # /* run VM compactor after process kill */
    BSD_MEMSTAT_FREEZE_DISABLE = 17  # /* disable freeze and kill frozen processes */
    BSD_MEMSTAT_RELAUNCH_FLAGS = 18  # /* flags representing jetsam behavior; based on launchd data */


class DBG_BSD_KEVENT(enum.Enum):
    BSD_KEVENT_KQ_PROCESS_BEGIN = 1
    BSD_KEVENT_KQ_PROCESS_END = 2
    BSD_KEVENT_KQWQ_PROCESS_BEGIN = 3
    BSD_KEVENT_KQWQ_PROCESS_END = 4
    BSD_KEVENT_KQWQ_BIND = 5
    BSD_KEVENT_KQWQ_UNBIND = 6
    BSD_KEVENT_KQWQ_THREQUEST = 7
    BSD_KEVENT_KQWL_PROCESS_BEGIN = 8
    BSD_KEVENT_KQWL_PROCESS_END = 9
    BSD_KEVENT_KQWL_THREQUEST = 10
    BSD_KEVENT_KQWL_THADJUST = 11
    BSD_KEVENT_KQ_REGISTER = 12
    BSD_KEVENT_KQWQ_REGISTER = 13
    BSD_KEVENT_KQWL_REGISTER = 14
    BSD_KEVENT_KNOTE_ACTIVATE = 15
    BSD_KEVENT_KQ_PROCESS = 16
    BSD_KEVENT_KQWQ_PROCESS = 17
    BSD_KEVENT_KQWL_PROCESS = 18
    BSD_KEVENT_KQWL_BIND = 19
    BSD_KEVENT_KQWL_UNBIND = 20
    BSD_KEVENT_KNOTE_ENABLE = 21
    BSD_KEVENT_KNOTE_VANISHED = 22


class DBG_TRACE(enum.Enum):
    DBG_TRACE_DATA = 0
    DBG_TRACE_STRING = 1
    DBG_TRACE_INFO = 2


class DBG_CORESTORAGE(enum.Enum):
    DBG_CS_IO = 0


class DBG_SECURITY(enum.Enum):
    DBG_SEC_KERNEL = 0  # /* raw entropy collected by the kernel */
    DBG_SEC_SANDBOX = 1


class DBG_MONOTONIC(enum.Enum):
    DBG_MT_INSTRS_CYCLES = 1
    DBG_MT_DEBUG = 2
    DBG_MT_RESOURCES_PROC_EXIT = 3
    DBG_MT_RESOURCES_THR_EXIT = 4
    DBG_MT_TMPTH = 0xfe
    DBG_MT_TMPCPU = 0xff


class DBG_MISC(enum.Enum):
    DBG_MISC_COREBRIGHTNESS = 0x01
    DBG_MISC_VIDEOENG = 0x02
    DBG_EVENT = 0x10
    DBG_MISC_INSTRUMENTS = 0x11
    DBG_MISC_INSTRUMENTSBT = 0x12
    DBG_MISC_LAYOUT = 0x1a
    DBG_BUFFER = 0x20


class DBG_DYLD(enum.Enum):
    DBG_DYLD_UUID = 5


class DBG_DYLD_UUID(enum.Enum):
    DBG_DYLD_UUID_MAP_A = (0)
    DBG_DYLD_UUID_MAP_B = (1)
    DBG_DYLD_UUID_MAP_32_A = (2)
    DBG_DYLD_UUID_MAP_32_B = (3)
    DBG_DYLD_UUID_MAP_32_C = (4)
    DBG_DYLD_UUID_UNMAP_A = (5)
    DBG_DYLD_UUID_UNMAP_B = (6)
    DBG_DYLD_UUID_UNMAP_32_A = (7)
    DBG_DYLD_UUID_UNMAP_32_B = (8)
    DBG_DYLD_UUID_UNMAP_32_C = (9)
    DBG_DYLD_UUID_SHARED_CACHE_A = (10)
    DBG_DYLD_UUID_SHARED_CACHE_B = (11)
    DBG_DYLD_UUID_SHARED_CACHE_32_A = (12)
    DBG_DYLD_UUID_SHARED_CACHE_32_B = (13)
    DBG_DYLD_UUID_SHARED_CACHE_32_C = (14)
    DBG_DYLD_AOT_UUID_MAP_A = (15)
    DBG_DYLD_AOT_UUID_MAP_B = (16)


class DBG_DKRW(enum.Enum):
    DKIO_DONE = 0x01
    DKIO_READ = 0x02
    DKIO_ASYNC = 0x04
    DKIO_META = 0x08
    DKIO_PAGING = 0x10
    DKIO_THROTTLE = 0x20  # /* Deprecated, still provided so fs_usage doesn't break */
    DKIO_PASSIVE = 0x40
    DKIO_NOCACHE = 0x80
    DKIO_TIER_MASK = 0xF00
    DKIO_TIER_SHIFT = 8
    DKIO_TIER_UPGRADE = 0x1000


class DBG_APPS(enum.Enum):
    DBG_APP_LOGINWINDOW = 0x03
    DBG_APP_AUDIO = 0x04
    DBG_APP_SYSTEMUI = 0x05
    DBG_APP_SIGNPOST = 0x0A
    DBG_APP_APPKIT = 0x0C
    DBG_APP_UIKIT = 0x0D
    DBG_APP_DFR = 0x0E
    DBG_APP_LAYOUT = 0x0F
    DBG_APP_COREDATA = 0x10
    DBG_APP_SAMBA = 0x80
    DBG_APP_EOSSUPPORT = 0x81
    DBG_APP_MACEFIMANAGER = 0x82


class DBG_THROTTLE(enum.Enum):
    OPEN_THROTTLE_WINDOW = 0x1
    PROCESS_THROTTLED = 0x2
    IO_THROTTLE_DISABLE = 0x3
    IO_TIER_UPL_MISMATCH = 0x4


class DBG_PERF(enum.Enum):
    PERF_EVENT = 0
    PERF_DATA = 1
    PERF_STK = 2


class DBG_IMPORTANCE(enum.Enum):
    IMP_ASSERTION = 0x10  # /* Task takes/drops a boost assertion */
    IMP_BOOST = 0x11  # /* Task boost level changed */
    IMP_MSG = 0x12  # /* boosting message sent by donating task on donating port */
    IMP_WATCHPORT = 0x13  # /* port marked as watchport, and boost was transferred to the watched task */
    IMP_TASK_SUPPRESSION = 0x17  # /* Task changed suppression behaviors */
    IMP_TASK_APPTYPE = 0x18  # /* Task launched with apptype */
    IMP_UPDATE = 0x19  # /* Requested -> effective calculation */
    IMP_USYNCH_QOS_OVERRIDE = 0x1A  # /* Userspace synchronization applied QoS override to resource owning thread */
    IMP_DONOR_CHANGE = 0x1B  # /* The iit_donor bit changed */
    IMP_MAIN_THREAD_QOS = 0x1C  # /* The task's main thread QoS was set */
    IMP_SYNC_IPC_QOS = 0x1D  # /* Sync IPC QOS override */
    IMP_TASK_POLICY_DARWIN_BG = 0x21
    IMP_TASK_POLICY_IOPOL = 0x22
    IMP_TASK_POLICY_IO = 0x23
    IMP_TASK_POLICY_PASSIVE_IO = 0x24
    IMP_TASK_POLICY_DARWIN_BG_IOPOL = 0x27
    IMP_TASK_POLICY_TAL = 0x28
    IMP_TASK_POLICY_BOOST = 0x29
    IMP_TASK_POLICY_ROLE = 0x2A
    IMP_TASK_POLICY_TERMINATED = 0x2C
    IMP_TASK_POLICY_NEW_SOCKETS_BG = 0x2D
    IMP_TASK_POLICY_SUP_ACTIVE = 0x2E
    IMP_TASK_POLICY_LATENCY_QOS = 0x2F
    IMP_TASK_POLICY_THROUGH_QOS = 0x30
    IMP_TASK_POLICY_WATCHERS_BG = 0x31
    IMP_TASK_POLICY_SFI_MANAGED = 0x34
    IMP_TASK_POLICY_ALL_SOCKETS_BG = 0x37
    IMP_TASK_POLICY_BASE_LATENCY_AND_THROUGHPUT_QOS = 0x39  # /* latency as value1, throughput as value2 */
    IMP_TASK_POLICY_OVERRIDE_LATENCY_AND_THROUGHPUT_QOS = 0x3A  # /* latency as value1, throughput as value2 */
    IMP_TASK_POLICY_PIDBIND_BG = 0x32
    IMP_TASK_POLICY_QOS_OVERRIDE = 0x36
    IMP_TASK_POLICY_QOS_AND_RELPRIO = 0x38  # /* QoS as value1, relative priority as value2 */
    IMP_TASK_POLICY_QOS_WORKQ_OVERRIDE = 0x3B
    IMP_TASK_POLICY_QOS_PROMOTE = 0x3C
    IMP_TASK_POLICY_QOS_KEVENT_OVERRIDE = 0x3D
    IMP_TASK_POLICY_QOS_SERVICER_OVERRIDE = 0x3E


class IMP_ASSERTION(enum.Enum):
    IMP_HOLD = 0x2  # /* Task holds a boost assertion */
    IMP_DROP = 0x4  # /* Task drops a boost assertion */
    IMP_EXTERN = 0x8  # /* boost assertion moved from kernel to userspace responsibility (externalized) */


class IMP_BOOST(enum.Enum):
    IMP_BOOSTED = 0x1
    IMP_UNBOOSTED = 0x2  # /* Task drops a boost assertion */


class IMP_MSG(enum.Enum):
    IMP_MSG_SEND = 0x1  # /* boosting message sent by donating task on donating port */
    IMP_MSG_DELV = 0x2  # /* boosting message delivered to task */


class IMP_UPDATE(enum.Enum):
    IMP_UPDATE_TASK_CREATE = 0x1


class IMP_USYNCH_QOS_OVERRIDE(enum.Enum):
    IMP_USYNCH_ADD_OVERRIDE = 0x0  # /* add override for a contended resource */
    IMP_USYNCH_REMOVE_OVERRIDE = 0x1  # /* remove override for a contended resource */


class IMP_DONOR_CHANGE(enum.Enum):
    IMP_DONOR_UPDATE_LIVE_DONOR_STATE = 0x0
    IMP_DONOR_INIT_DONOR_STATE = 0x1


class IMP_SYNC_IPC_QOS(enum.Enum):
    IMP_SYNC_IPC_QOS_APPLIED = 0x0
    IMP_SYNC_IPC_QOS_REMOVED = 0x1
    IMP_SYNC_IPC_QOS_OVERFLOW = 0x2
    IMP_SYNC_IPC_QOS_UNDERFLOW = 0x3


class DBG_TURNSTILE(enum.Enum):
    TURNSTILE_HEAP_OPERATIONS = 0x10
    TURNSTILE_PRIORITY_OPERATIONS = 0x20
    TURNSTILE_FREELIST_OPERATIONS = 0x30


class TURNSTILE_HEAP_OPERATIONS(enum.Enum):
    THREAD_ADDED_TO_TURNSTILE_WAITQ = 0x1
    THREAD_REMOVED_FROM_TURNSTILE_WAITQ = 0x2
    THREAD_MOVED_IN_TURNSTILE_WAITQ = 0x3
    TURNSTILE_ADDED_TO_TURNSTILE_HEAP = 0x4
    TURNSTILE_REMOVED_FROM_TURNSTILE_HEAP = 0x5
    TURNSTILE_MOVED_IN_TURNSTILE_HEAP = 0x6
    TURNSTILE_ADDED_TO_THREAD_HEAP = 0x7
    TURNSTILE_REMOVED_FROM_THREAD_HEAP = 0x8
    TURNSTILE_MOVED_IN_THREAD_HEAP = 0x9
    TURNSTILE_UPDATE_STOPPED_BY_LIMIT = 0xa
    THREAD_NOT_WAITING_ON_TURNSTILE = 0xb


class TURNSTILE_PRIORITY_OPERATIONS(enum.Enum):
    TURNSTILE_PRIORITY_CHANGE = 0x1
    THREAD_USER_PROMOTION_CHANGE = 0x2


class TURNSTILE_FREELIST_OPERATIONS(enum.Enum):
    TURNSTILE_PREPARE = 0x1
    TURNSTILE_COMPLETE = 0x2


class DBG_BANK(enum.Enum):
    BANK_ACCOUNT_INFO = 0x10  # /* Trace points related to bank account struct */
    BANK_TASK_INFO = 0x11  # /* Trace points related to bank task struct */


class DBG_ATM(enum.Enum):
    ATM_SUBAID_INFO = 0x10
    ATM_GETVALUE_INFO = 0x20
    ATM_UNREGISTER_INFO = 0x30


class BANK_ACCOUNT_INFO(enum.Enum):
    BANK_SETTLE_CPU_TIME = 0x1  # /* Bank ledger(chit) rolled up to tasks. */
    BANK_SECURE_ORIGINATOR_CHANGED = 0x2  # /* Secure Originator changed. */
    BANK_SETTLE_ENERGY = 0x3  # /* Bank ledger(energy field) rolled up to tasks. */


class ATM_SUBAID_INFO(enum.Enum):
    ATM_MIN_CALLED = 0x1
    ATM_LINK_LIST_TRIM = 0x2


class ATM_GETVALUE_INFO(enum.Enum):
    ATM_VALUE_REPLACED = 0x1
    ATM_VALUE_ADDED = 0x2


class ATM_UNREGISTER_INFO(enum.Enum):
    ATM_VALUE_UNREGISTERED = 0x1
    ATM_VALUE_DIFF_MAILBOX = 0x2


class DBG_DAEMON(enum.Enum):
    DBG_DAEMON_COREDUET = 0x1
    DBG_DAEMON_POWERD = 0x2




kd_threadmap = Struct(
    'tid' / Int64ul,
    'pid' / Int32ul,
    'process' / FixedSized(0x14, CString('utf8')),
)

kd_header_v2 = Struct(
    'tag' / Int32ul,
    'number_of_treads' / Int32ul,
    Padding(12),
    'is_64bit' / Int32ul,
    'tick_frequency' / Int64ul,
    Padding(0x100),
    'threadmap' / Array(lambda ctx: ctx.number_of_treads, kd_threadmap),
    '_pad' / GreedyRange(Const(0, Byte)),
)


class BplistAdapter(Adapter):
    def _decode(self, obj, context, path):
        return plistlib.loads(obj)

    def _encode(self, obj, context, path):
        return plistlib.dumps(obj)


class KCdataParse(Adapter):

    def _decode(self, obj, context, path):
        return kc_data_parse(obj)


kd_header_v3 = Struct(
    'tag' / Int32ul,
    'sub_tag' / Int32ul,
    'length' / Int64ul,
    'timebase_numer' / Int32ul,
    'timebase_denom' / Int32ul,
    'timestamp' / Int64ul,
    'walltime_secs' / Int64ul,
    'walltime_usecs' / Int32ul,
    'timezone_minuteswest' / Int32ul,
    'timezone_dst' / Int32ul,
    'flags' / Int32ul,
)

CLASS_DICT = vars()

class KdBufParser:

    def __init__(self, timestamp, args_buf, tid, debug_id, cpuid, unused):
        self.timestamp = timestamp
        self.args = struct.unpack('<QQQQ', args_buf)
        self.tid = tid
        self.cpuid = cpuid
        self.unused = unused
        self.debug_id = debug_id
        self.event_id = debug_id & KDBG_EVENTID_MASK
        self.func_code = debug_id & KDBG_FUNC_MASK
        self.class_code = kdbg_extract_class(debug_id)
        self.subclass_code = kdbg_extract_subclass(debug_id)
        self.final_code = kdbg_extract_code(debug_id)

    @classmethod
    def decode(cls, buf_io: io.BytesIO):
        while True:
            buf = buf_io.read(64)
            if not buf:
                return
            yield cls(*struct.unpack(KD_BUF_FORMAT, buf))


def _format_class(classes, code):
    if classes:
        try:
            classes_name = classes(code).name
            return classes_name, f'{classes_name:<18}'
        except ValueError:
            return None, f'''{'Error(' + (str(code)) + ")"}'''
    else:
        return None, f'''{'Error(' + (str(code)) + ")"}'''


SSHOT_TAG = 0x8002
IMAGES_TAG = 0x8004
KERNEL_EXTENSIONS_TAG = 0x8005
CONFIG_TAG = 0x8006
KERNEL_TAG = 0x8008
MACHINE_TAG = 0x8c00
CPU_EVENTS_TAG = 0x8c01
CPU_EVENTS_TAG2 = 0x8013
CPU_EVENTS_NULL = 0x800e
RAW_VERSION3 = 0x00001000
V3_NULL_CHUNK = 0x00002000
V3_CONFIG = 0x00001b00
V3_CPU_HEADER_TAG = 0x00001c00
V3_THREAD_MAP = 0x00001d00
V3_RAW_EVENTS = 0x00001e00

kd_cpumap = Struct(
    'cpu_id' / Int32ul,
    'flags' / Int32ul,
    'name' / FixedSized(8, CString('utf8')),
    'args' / Array(6, Int32ul),
)

kd_cpumap_header = Struct(
    'version_no' / Int32ul,
    'cpu_count' / Int32ul,
)

ktrace_pack = Struct(
    'tag' / Int32ul,
    'major' / Int16ul,
    'minor' / Int16ul,
    'length' / Int64ul,
    'data' / Switch(this.tag,
                    {MACHINE_TAG: FixedSized(this.length, BplistAdapter(GreedyBytes)),
                     CONFIG_TAG: FixedSized(this.length, BplistAdapter(GreedyBytes)),
                     KERNEL_TAG: FixedSized(this.length, BplistAdapter(GreedyBytes)),
                     KERNEL_EXTENSIONS_TAG: FixedSized(this.length, BplistAdapter(GreedyBytes)),
                     IMAGES_TAG: FixedSized(this.length, BplistAdapter(GreedyBytes)),
                     SSHOT_TAG: FixedSized(this.length, KCdataParse(GreedyBytes)),
                     CPU_EVENTS_TAG: FixedSized(this.length, GreedyBytes),
                     CPU_EVENTS_TAG2: FixedSized(this.length, GreedyBytes),
                     CPU_EVENTS_NULL: FixedSized(this.length, GreedyBytes),
                     V3_NULL_CHUNK: FixedSized(this.length, GreedyBytes),
                     V3_RAW_EVENTS: FixedSized(this.length, GreedyBytes),
                     V3_CPU_HEADER_TAG: kd_cpumap_header,
                     V3_THREAD_MAP: FixedSized(this.length, GreedyRange(kd_threadmap))},
                    default=GreedyBytes),
)


class MyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, bytes):
            return str(obj)
        return json.JSONEncoder.default(self, obj)


class KperfData:
    def __init__(self, traceCodesFile={}, filter_pid=None, filter_process=None):
        self.machine_info = None
        self.cpu_header = None
        self.kd_cpu_map = None
        self.report_config = None
        self.ktrace_config = None
        self.kd_header_v3 = None
        self.stack_shot = None
        self.trace_codes = traceCodesFile
        self.threads_pids = {}
        self.version = None
        self.filter_tid = filter_pid
        self.filter_process = filter_process

    def set_threads_pids(self, threads):
        for thread in threads:
            self.threads_pids[thread.tid] = (thread.pid, thread.process)

    def _format_process(self, tid):
        pid, process_name = self.threads_pids.get(tid, (None, None))
        return pid, process_name, f'{process_name}({pid})' if pid else f'Error: tid {tid}'

    def parse(self, kd_buf: bytes):
        buf_io = io.BytesIO(kd_buf)
        version = kd_buf[:4]
        if version == b'\x07X\xa2Y':
            yield kc_data_parse(buf_io.read())
        elif version == VERSION2_FLAG:
            for event in self.parse_v2(buf_io):
                yield event
        elif version == VERSION3_FLAG:
            for event in self.parse_v3(buf_io):
                yield event
        else:
            for event in KdBufParser.decode(buf_io):
                yield event

    def parse_v2(self, buf_io: io.BytesIO):
        parsed_header = kd_header_v2.parse_stream(buf_io)
        self.set_threads_pids(parsed_header.threadmap)
        for event in KdBufParser.decode(buf_io):
            yield event

    def parse_v3(self, buf_io: io.BytesIO):
        self.kd_header_v3 = kd_header_v3.parse_stream(buf_io)
        while True:
            try:
                raw_events = ktrace_pack.parse_stream(buf_io)
            except StreamError:
                break
            if raw_events.tag in (V3_NULL_CHUNK, CPU_EVENTS_NULL):
                continue
            elif raw_events.tag == CONFIG_TAG:
                self.ktrace_config = raw_events
                GreedyRange(Const(0, Byte)).parse_stream(buf_io)
            elif raw_events.tag == KERNEL_TAG:
                self.report_config = raw_events
                GreedyRange(Const(0, Byte)).parse_stream(buf_io)
            elif raw_events.tag == SSHOT_TAG:
                self.stack_shot = raw_events
            elif raw_events.tag == V3_THREAD_MAP:
                self.set_threads_pids(raw_events.data)
            elif raw_events.tag == MACHINE_TAG:
                self.machine_info = raw_events
                GreedyRange(Const(0, Byte)).parse_stream(buf_io)
            elif raw_events.tag == V3_CPU_HEADER_TAG:
                self.cpu_header = raw_events
                self.kd_cpu_map = FixedSized(raw_events.length - kd_cpumap_header.sizeof(),
                                             GreedyRange(kd_cpumap)).parse_stream(buf_io)
            elif raw_events.tag == V3_RAW_EVENTS:
                events_buf = io.BytesIO(raw_events.data)
                events_buf.read(8)  # Future events timestamp struct.pack('Q', 0)
                for event in KdBufParser.decode(events_buf):
                    yield event

    def to_dict(self, kd_buf):
        for event in self.parse(kd_buf):
            yield event

    def to_str(self, kd_buf: bytes):
        for event in self.parse(kd_buf):
            if not isinstance(event, KdBufParser):
                yield event
                return
            pid, process_name, process_str = self._format_process(event.tid)
            if self.filter_tid and self.filter_tid != pid:
                continue
            if self.filter_process and self.filter_process != process_name:
                continue
            formatted_data = ''
            formatted_data += f'{process_str:<27}'
            if event.event_id in self.trace_codes:
                name = self.trace_codes[event.event_id] + f' ({hex(event.event_id)})'
            else:
                name = hex(event.event_id)

            formatted_data += f'{name:<60}'
            classes_name, _str = _format_class(DebugClasses, event.class_code)
            formatted_data += f'{_str:<18}'
            classes_name, _str = _format_class(CLASS_DICT.get(classes_name), event.subclass_code)
            formatted_data += f'{_str:<30}'
            try:
                formatted_data += f'{DgbFuncQual(event.func_code).name:<15}'
            except ValueError:
                formatted_data += f'{"Error":<16}'
            yield formatted_data



import logging
import struct

from construct import Struct, PascalString, Int, Optional

from ios_device.util.lockdown import LockdownClient

LOCATION_STRUCT = Struct(
    status=Int,
    latitude=Optional(PascalString(Int, "utf8")),
    longitude=Optional(PascalString(Int, "utf8")),
)


class SimulateLocation(object):
    SERVICE_NAME = 'com.apple.dt.simulatelocation'

    def __init__(self, lockdown=None, udid=None, network=None, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.lockdown = lockdown or LockdownClient(udid=udid, network=network)
        self.service = self.lockdown.start_service(self.SERVICE_NAME)

    def clear(self):
        """ stop simulation """
        self.service.sock.sendall(LOCATION_STRUCT.build({'status': 1}))

    def set(self, latitude: float, longitude: float):
        """ set simulation """
        location_byte = LOCATION_STRUCT.build({
            'status': 0,
            'latitude': str(latitude),
            'longitude': str(longitude)
        })
        self.service.sock.sendall(location_byte)



"""
启动 app
"""
import os
import sys

sys.path.append(os.getcwd())

from ios_device.servers.Instrument import InstrumentServer


def _launch_app(rpc, bundleid):
    rpc._start()

    def on_channel_message(res):
        print(res.auxiliaries, res.selector)

    channel = "com.apple.instruments.server.services.processcontrol"
    rpc.register_channel_callback(channel, on_channel_message)
    pid = rpc.call(channel, "launchSuspendedProcessWithDevicePath:bundleIdentifier:environment:arguments:options:", "",
                   bundleid, {}, [], {"StartSuspendedKey": 0, "KillExisting": 1}).selector
    print("start", pid)


if __name__ == '__main__':
    rpc = InstrumentServer().init()
    _launch_app(rpc, 'cn.rongcloud.im')
    rpc.stop()
